##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for flowsheet_model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, Constraint, Set, Var
from pyomo.common.config import ConfigBlock
from idaes.core import (declare_process_block_class, PhysicalParameterBlock,
                        StateBlock, StateBlockData)
from idaes.core.util.exceptions import (PropertyPackageError,
                                        PropertyNotSupportedError)

# -----------------------------------------------------------------------------
# Test ParameterBlock
@declare_process_block_class("ParameterBlock")
class _ParameterBlock(PhysicalParameterBlock):
    pass


def test_config_block():
    # Test that PhysicalParameterBlock gets module information
    m = ConcreteModel()
    m.p = ParameterBlock()

    assert len(m.p.config) == 1
    assert isinstance(m.p.config.default_arguments, ConfigBlock)
    assert len(m.p.config.default_arguments) == 0


def test_PhysicalParameterBlock():
    # Test that PhysicalParameterBlock builds correctly
    m = ConcreteModel()
    m.p = ParameterBlock()
    super(_ParameterBlock, m.p).build()


def test_PhysicalParameter_NotImplementedErrors():
    # Test that class methods return NotImplementedError
    m = ConcreteModel()
    m.p = ParameterBlock()

    with pytest.raises(NotImplementedError):
        m.p.get_metadata()


# -----------------------------------------------------------------------------
# Test StateBlock
@declare_process_block_class("TestStateBlock", block_class=StateBlock)
class _StateBlockData(StateBlockData):
    pass


@declare_process_block_class("TestStateBlock2", block_class=StateBlock)
class _StateBlockData(StateBlockData):
    def define_state_vars(self):
        return {}


def test_StateBlockBase_initialize():
    # Test that StateBlock initialize method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[1, 2])
    m.p = TestStateBlock2(m.s)

    with pytest.raises(NotImplementedError):
        m.p.initialize()


def test_StateBlockBase_report():
    # Test that StateBlock initialize method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[0, 1])
    m.p = TestStateBlock2(m.s)

    m.p.report(dof=True)


# -----------------------------------------------------------------------------
# Test StateBlockData
def test_StateBlock_config():
    # Test that StateBlockData config has correct arguments
    m = ConcreteModel()
    m.p = TestStateBlock()

    assert len(m.p.config) == 3
    assert hasattr(m.p.config, "has_phase_equilibrium")
    assert hasattr(m.p.config, "defined_state")
    assert hasattr(m.p.config, "parameters")

    m.p.config.has_phase_equilibrium = True
    m.p.config.has_phase_equilibrium = False
    with pytest.raises(ValueError):
        m.p.config.has_phase_equilibrium = 'foo'
    with pytest.raises(ValueError):
        m.p.config.has_phase_equilibrium = 10

    m.p.config.defined_state = True
    m.p.config.defined_state = False
    with pytest.raises(ValueError):
        m.p.config.defined_state = 'foo'
    with pytest.raises(ValueError):
        m.p.config.defined_state = 10


def test_StateBlock_NotImplementedErrors():
    # Test that placeholder methods return NotImplementedErrors
    m = ConcreteModel()
    m.p = TestStateBlock()

    with pytest.raises(NotImplementedError):
        m.p.define_state_vars()
    with pytest.raises(NotImplementedError):
        m.p.define_port_members()
    with pytest.raises(NotImplementedError):
        m.p.get_material_flow_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_material_density_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_material_diffusion_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_enthalpy_flow_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_energy_density_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_energy_diffusion_terms()
    with pytest.raises(NotImplementedError):
        m.p.calculate_bubble_point_temperature()
    with pytest.raises(NotImplementedError):
        m.p.calculate_dew_point_temperature()
    with pytest.raises(NotImplementedError):
        m.p.calculate_bubble_point_pressure()
    with pytest.raises(NotImplementedError):
        m.p.calculate_dew_point_pressure()

# -----------------------------------------------------------------------------
# Test parameter block reference attribute
@declare_process_block_class("Parameters")
class _Parameters(PhysicalParameterBlock):
    def build(self):
        super(_Parameters, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'a': {'method': 'a_method'},
                            'recursion1': {'method': '_recursion1'},
                            'recursion2': {'method': '_recursion2'},
                            'not_callable': {'method': 'test_obj'},
                            'raise_exception': {'method': '_raise_exception'},
                            'not_supported': {'method': False},
                            'does_not_create_component': {
                                'method': '_does_not_create_component'}})


@declare_process_block_class("StateTest", block_class=StateBlock)
class _StateTest(StateBlockData):
    def build(self):
        super(_StateTest, self).build()


def test_param_ref():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = StateTest(default={"parameters": m.pb})

    assert m.p._params == m.p.config.parameters


# -----------------------------------------------------------------------------
# Test properties __getattr__ method
@declare_process_block_class("State", block_class=StateBlock)
class _State(StateBlockData):
    def build(self):
        super(StateBlockData, self).build()

        self.test_obj = 1

    def a_method(self):
        self.a = Var(initialize=1)

    def _recursion1(self):
        self.recursive_cons1 = Constraint(expr=self.recursion2 == 1)

    def _recursion2(self):
        self.recursive_cons2 = Constraint(expr=self.recursion1 == 1)

    def _raise_exception(self):
        raise Exception()

    def _does_not_create_component(self):
        pass


@pytest.fixture()
def m():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = State(default={"parameters": m.pb})

    return m


def test_getattr_add_var(m):
    assert isinstance(m.p.a, Var)
    assert m.p.a.value == 1


def test_getattr_protected(m):
    with pytest.raises(PropertyPackageError):
        # Call a protected component that does not exist
        m.p.cons = Constraint(expr=m.p._foo == 1)


def test_getattr_recursion(m):
    with pytest.raises(PropertyPackageError):
        # Call a component that triggers a recursive loop of calls
        m.p.cons = Constraint(expr=m.p.recursion1 == 1)


def test_getattr_does_not_exist(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.does_not_exist == 1)


def test_getattr_not_callable(m):
    with pytest.raises(PropertyPackageError):
        m.p.cons = Constraint(expr=m.p.not_callable == 1)


def test_getattr_not_supported(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.not_supported == 1)


def test_getattr_raise_exception(m):
    with pytest.raises(Exception):
        m.p.cons = Constraint(expr=m.p.raise_exception == 1)


# TODO : Need a test for cases where method does not create property
#def test_getattr_does_not_create_component(m):
#    with pytest.raises(PropertyPackageError):
#        m.p.cons = Constraint(expr=m.p.does_not_create_component == 1)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""Commandline Utilities for Managing the IDAES Data Directory"""

__author__ = "John Eslick"

import os
import idaes
import click
from idaes.commands import cb

@cb.command(name="data-directory", help="Show IDAES data directory")
@click.option("--exists", is_flag=True, help="Show if the directory exists")
@click.option("--create", is_flag=True, help="Create the directory")
def data_directory(exists, create):
    if create:
        click.echo("Creating {}".format(idaes.data_directory))
        idaes._create_data_dir()
    elif exists:
        click.echo(os.path.exists(idaes.data_directory))
    else:
        click.echo(idaes.data_directory)

@cb.command(name="bin-directory", help="Show IDAES executable file directory")
@click.option("--exists", is_flag=True, help="Show if the directory exists")
@click.option("--create", is_flag=True, help="Create the directory")
def bin_directory(exists, create):
    if create:
        click.echo("Creating {}".format(idaes.bin_directory))
        idaes._create_bin_dir()
    elif exists:
        click.echo(os.path.exists(idaes.bin_directory))
    else:
        click.echo(idaes.bin_directory)

@cb.command(name="lib-directory", help="Show IDAES library file directory")
@click.option("--exists", is_flag=True, help="Show if the directory exists")
@click.option("--create", is_flag=True, help="Create the directory")
def lib_directory(exists, create):
    if create:
        click.echo("Creating {}".format(idaes.lib_directory))
        idaes._create_lib_dir()
    elif exists:
        click.echo(os.path.exists(idaes.lib_directory))
    else:
        click.echo(idaes.lib_directory)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# This subroutine is no longer used in the ripe-pyomo implementation


def writeripe(aterm, ccon):
    import os
    import numpy as np
    import ripe

    data = ripe.data

    # wrapper for solving RIPE problems
    ndata = data["otherdata"]["npc"] * max(data["otherdata"]["ntimedata"], 1)
    if ccon < 0:
        clist = range(1, min(ndata, len(data["otherdata"]["slabels"])))
        [aterm, kils] = writemodel(aterm, ccon)
        if kils == 1:
            os.system("gams ripe.gms")
            f = open("rates.put")
            lf = f.read()
            f.close()
            lf2 = lf.split("\n")
            lf2 = lf2[0].split(" ")
            lf2 = [float(x) for x in lf2 if x != ""]
            sigmaest = [0] * len(lf2)
            bic = [0] * int(ndata)
            for i in range(len(lf2)):
                sigmaest[i] = lf2[i] / (ndata - 1)
        else:
            exit
    else:
        clist = list([ccon])
    for ccon in clist:
        [aterm, kils] = writemodel(aterm, ccon)
        if kils == 1:
            os.system("gams ripe.gms")
            f = open("rates.put")
            lf = f.read()
            f.close()
            lf2 = lf.split("\n")
            lf2 = lf2[0].split(" ")
            lf2 = [float(x) for x in lf2 if x != ""]
            sigmaest = [0] * len(lf2)
            bic = [0] * int(ndata)
            for i in range(len(lf2)):
                sigmaest[i] = lf2[i] / (ndata - 1)
            bic[ccon] = (
                sum(np.divide(lf2, sigmaest))
                + np.log(data["otherdata"]["npc"] * len(data["otherdata"]["slabels"]))
                * ccon
            )
            # ripeci(aterm, ccon)
        else:
            exit


def writemodel(aterm, ccon):
    import numpy as np
    import os
    import sys
    from idaes.surrogate import alamopy
    import ripe
    from alamopy import deletefile

    data, rspace = ripe.data, ripe.rspace

    # First an activity term matrix is defined to make assignment easier later
    if "clc" in data["otherdata"]:
        aterm = np.zeros(
            [
                data["otherdata"]["npc"],
                data["otherdata"]["ntimedata"],
                data["rxnet"]["mechanisms"],
            ]
        )
        for i in range(data["otherdata"]["npc"]):
            for j in range(data["otherdata"]["ntimedata"]):
                # t is an index on the current mechanism
                t = 0
                if "clc" in data["otherdata"]:
                    for m in range(len(data["otherdata"]["clc"])):
                        mech = data["otherdata"]["clc"][m]
                        # Check for CLC mechanisms
                        if mech == "c1":
                            aterm[i, j, t] = (
                                1.0 - data["concdata"]["X"][i][j]
                            )  # random nucleation
                        if mech == "c2":
                            aterm[i, j, t] = (
                                2.0
                                / 3.0
                                * data["concdata"]["X"][i][j] ** (2.0 / 3.0 - 3.0 / 2.0)
                            )  # Power law of oer 2/3
                        if mech == "c3":
                            aterm[i, j, t] = 2.0 * data["concdata"]["X"][i][j] ** (
                                2.0 - 1.0 / 2.0
                            )  # Power law model of order 2
                        if mech == "c4":
                            aterm[i, j, t] = 4.0 * data["concdata"]["X"][i][j] ** (
                                4.0 - 1.0 / 4.0
                            )  # power law model of order 4
                        if mech == "c5":
                            aterm[i, j, t] = 3.0 * data["concdata"]["X"][i][j] ** (
                                3.0 - 1.0 / 3.0
                            )  # power law model of order 4
                        if mech == "c6":
                            aterm[i, j, t] = 1.0  # power law model of order 1
                        if mech == "c7":
                            aterm[i, j, t] = (
                                2.0
                                * (1.0 - data["concdata"]["X"][i][j])
                                * np.power(
                                    -np.log(1.0 - data["concdata"]["X"][i][j]), 1.5
                                )
                            )  # nucleation mode - Avrami-Erofeev eq of order 2
                        if mech == "c8":
                            aterm[i, j, t] = (
                                3.0
                                * (1.0 - data["concdata"]["X"][i][j])
                                * np.power(
                                    -np.log(1.0 - data["concdata"]["X"][i][j]),
                                    3.0 - (1.0 / 3.0),
                                )
                            )  # '' order 3
                        if mech == "c9":
                            aterm[i, j, t] = (
                                0.5
                                * (1.0 - data["concdata"]["X"][i][j])
                                * (-np.log(1.0 - data["concdata"]["X"][i][j]))
                                ** (0.5 - 1.0 / 0.5)
                            )  # '' order 0.5
                        if mech == "c10":
                            aterm[i, j, t] = (
                                1.5
                                * (1.0 - data["concdata"]["X"][i][j])
                                * np.power(
                                    -np.log(1.0 - data["concdata"]["X"][i][j]),
                                    1.5 - (1.0 / 1.5),
                                )
                            )  # '' order 1.5
                        if mech == "c11":
                            aterm[i, j, t] = (
                                4.0
                                * (1.0 - data["concdata"]["X"][i][j])
                                * (-np.log(1.0 - data["concdata"]["X"][i][j]))
                                ** (4.0 - 1.0 / 4.0)
                            )  # '' order 4
                        if mech == "c12":
                            aterm[i, j, t] = data["concdata"]["X"][i][j] * (
                                1.0 - data["concdata"]["X"][i][j]
                            )  # Prout tompkins
                        # diffusion models
                        if mech == "c13":
                            aterm[i, j, t] = (
                                3.0
                                * (1.0 - data["concdata"]["X"][i][j]) ** (1.0 / 3.0)
                                * (
                                    2.0
                                    * (
                                        (1.0 - data["concdata"]["X"][i][j])
                                        ** (-1.0 / 3.0)
                                        - 1.0
                                    )
                                )
                                ** (-1)
                            )  # Jander equation (3d difsusion spherical sym)
                        if mech == "c14":
                            aterm[i, j, t] = (
                                3.0
                                / 2.0
                                * (1.0 + data["concdata"]["X"][i][j]) ** (2.0 / 3.0)
                                * (
                                    (
                                        (1.0 + data["concdata"]["X"][i][j])
                                        ** (-1.0 / 3.0)
                                        - 1.0
                                    )
                                )
                                ** (-1)
                            )  # antijander
                        if mech == "c15":
                            aterm[i, j, t] = 1.0 / (
                                -np.log(1.0 - data["concdata"]["X"][i][j])
                            )  # valensi
                        if mech == "c16":
                            aterm[i, j, t] = 1.0 / (
                                2.0 * data["concdata"]["X"][i][j]
                            )  # parabolic
                        if mech == "c17":
                            aterm[i, j, t] = (3.0 / 2.0) * (
                                (
                                    (1.0 - data["concdata"]["X"][i][j]) ** (-1.0 / 3.0)
                                    - 1.0
                                )
                            ) ** (
                                -1
                            )  # Ginstling-Brounstein eq 3d diff cylindrical
                        if mech == "c18":
                            aterm[i, j, t] = (
                                3.0
                                / 2.0
                                * (1.0 - data["concdata"]["X"][i][j]) ** (4.0 / 3.0)
                                * (
                                    (
                                        (1.0 - data["concdata"]["X"][i][j])
                                        ** (-1.0 / 3.0)
                                        - 1.0
                                    )
                                )
                                ** (-1)
                            )  # zhuralev, lesohkhin, tempelman
                        if mech == "c19":
                            aterm[i, j, t] = (1.0 - data["concdata"]["X"][i][j]) ** (
                                2.0 / 3.0
                            )
                        # if(mech=='c19'): aterm[i,j,t]=  (0.2)**(0.535) * 1.5 * (0.112)**(1.0/1.5)*((1.0)-data['concdata']['X'][i][j])*(-np.log(1.0-data['concdata']['X'][i][j]))**(1.0-1.0/1.5)                                                                                                                                                                      #                    if(mech=='c20'): aterm[i,j,t]=  (0.3)**(0.312) * 2.76 * (11.24)**(1.0/2.76)*((0.18 *1.479)-data['concdata']['X'][i][j])*(-np.log(1.0-data['concdata']['X'][i][j])/(0.18 *1.479))**(1.0-1.0/2.76)                                                                                                                                              #                    if(mech=='c21'): aterm[i,j,t]=  (0.3)**(2.13) * 1.63 * (28.4)**(1.0/1.63)*((0.135*1.479)-data['concdata']['X'][i][j])*(-np.log(1.0-data['concdata']['X'][i][j])/(0.135*1.479))**(1.0-1.0/1.63)
                        t = t + 1
    else:
        for m in range(data["rxnet"]["mechanisms"]):
            tempa = 1
            for s in data["otherdata"]["slabels"]:
                if data["rxnet"][i, s] < 0:
                    tempa = tempa * float(data["concdata"][s][i][j]) ** float(
                        abs(data["rxnet"][i, s])
                    )
                    aterm[i, j, m] = tempa

        for q in range(data["rxnet"]["mechanisms"]):
            if np.isinf(aterm[i, j, q]):
                aterm[i, j, q] = 0
                if np.isnan(aterm[i, j, q]):
                    aterm[i, j, q] = 0

        if "time" not in data["procdata"]:
            for q in range(data["rxnet"]["nrx"]):
                temp = 1
                for s in data["otherdata"]["slabels"]:
                    if data["rxnet"][q, s] < 0:
                        temp = temp * data["concdata"][s][i] ** abs(data["rxnet"][q, s])
                aterm[i, q] = temp

    # check for isothermal, single species observations
    if len(data["otherdata"]["slabels"]) == 1:
        if "T" not in data["otherdata"]["pclabels"]:
            sys.stdout = open(os.devnull, "w")
            r = alamopy.doalamo(
                np.asarray(aterm[0, :, :]),
                np.asarray(
                    data["ratedata"][str(data["otherdata"]["slabels"][0])][0, :]
                ),
                linfcns=1,
                constant=0,
                maxterms=1,
                expandoutput=1,
            )
            r = alamopy.almconfidence(r)
            sys.stdout = sys.__stdout__
            selected = r["model"].split(" * ")[-1]
            selected = selected[1:]
            coeff = r["model"].split(" * ")[0]
            coeff = coeff.split(" = ")[-1]
            cfi = r["conf_inv"][0].split(" : ")[-1]
            slist = [
                "Selected model is : " + rspace["clcnames"][int(selected) - 1],
                "Selected model form : " + rspace["clcforms"][int(selected) - 1],
                "Estimated rate constant : " + cfi,
            ]
            sys.stdout.write(slist[0] + "\n")
            sys.stdout.write(slist[1] + "\n")
            sys.stdout.write(slist[2] + "\n")
            with open("output.txt", "w") as o:
                o.write(slist[0] + "\n")
                o.write(slist[1] + "\n")
                o.write(slist[2] + "\n")
            deletefile("*logscratch*", "tempalm.py", "sv.alm", "tmpscratch")
            return [aterm, 0]
    else:
        writeminlp(aterm, ccon)
        return [aterm, 0]


def writeminlp(aterm, ccon):
    import ripe
    import numpy as np

    # Main ripe subroutine
    # initialize some stuff

    data, debug = ripe.data, ripe.debug

    ndatadic = {}
    for what in data["procdata"]:
        try:
            ndatadic[what] = len(data["procdata"][what])
        except Exception:
            ndatadic[what] = 1
        if "time" == what:
            ndatadic[what] = np.size(data["procdata"][what], 1)

    with open("ripe.gms", "w") as r:
        r.write("$offdigit\n")
        r.write("$offsymxref offsymlist\n")

        # write out species no matter what
        if "clc" in data["otherdata"]:
            r.write("set s /X/;\n")
        else:
            r.write("set s /a*" + data["otherdata"]["slabels"][-1] + "/;\n")
        r.write("alias(s,s1,s2,s3,s4);\n")

        # write npc set
        if data["otherdata"]["npc"] > 0:
            r.write("set pc /1*" + str(data["otherdata"]["npc"]) + "/;\n")

        # Is it isothermal?
        if "T" in data["procdata"]:
            r.write("parameters temp(pc);\n")
            if ndatadic["T"] > 1:
                #            r.write('set pc /1*'+str(ndatadic['T'])+'/;\n')
                for i in range(ndatadic["T"]):
                    r.write(
                        "temp('"
                        + str(i + 1)
                        + "')="
                        + str(data["procdata"]["T"][i])
                        + ";\n"
                    )
            else:
                r.write("temp('1')=" + str(data["procdata"]["T"]) + ";\n")
        if "time" in data["procdata"]:
            r.write("set time /1*" + str(ndatadic["time"]) + "/;\n")
            r.write("parameters t(pc,time);\n")
            for i in range(ndatadic["T"]):
                for j in range(ndatadic["time"]):
                    r.write(
                        "t('"
                        + str(i + 1)
                        + "','"
                        + str(j + 1)
                        + "')="
                        + str(data["procdata"]["time"][i][j])
                        + ";\n"
                    )

        # Some more equations and sets
        r.write("equations obj,cardcon,m1l,m1u,mc;\n")
        if "T" in data["otherdata"]["pclabels"]:
            r.write("equations m2l, m2u;\n")
        r.write("set m /1*" + str(data["rxnet"]["mechanisms"]) + "/;\n")
        r.write("set rx /1*" + str(data["rxnet"]["nrx"]) + "/;\n")

        if "clc" in data["otherdata"]:
            # iscomm = 0
            tstr = "set mclc(m) / "
            for m in data["otherdata"]["clc"]:
                tstr = tstr + str(data["gprxindex"][0][m])
                if m != data["otherdata"]["clc"][-1]:
                    tstr = tstr + ","
                else:
                    tstr = tstr + "/;\n"
            r.write(tstr)

        # Now we gotta write all those concentration values
        if data["otherdata"]["npc"] > 0:
            if "time" in data["procdata"]:
                r.write(
                    "parameters conc(pc,time,s),rate(pc,time,s),aterm(pc,time,m);\n"
                )
                #                r.write('parameters conc(pc,time,s),rate(pc,time,s),aterm(pc,time,rc);\n')
                for i in range(ndatadic["T"]):
                    for j in range(ndatadic["time"]):
                        for s in data["otherdata"]["slabels"]:
                            r.write(
                                "conc('"
                                + str(i + 1)
                                + "','"
                                + str(j + 1)
                                + "','"
                                + s
                                + "')="
                                + str(data["concdata"][s][i][j])
                                + ";\n"
                            )
                            r.write(
                                "rate('"
                                + str(i + 1)
                                + "','"
                                + str(j + 1)
                                + "','"
                                + s
                                + "')="
                                + str(data["ratedata"][s][i][j])
                                + ";\n"
                            )
                            for q in range(data["rxnet"]["mechanisms"]):
                                r.write(
                                    "aterm('"
                                    + str(i + 1)
                                    + "','"
                                    + str(j + 1)
                                    + "','"
                                    + str(q + 1)
                                    + "')="
                                    + str(aterm[i, j, q])
                                    + ";\n"
                                )

            else:
                r.write("parameters conc(pc,s),rate(pc,s);\n")
                for i in range(data["otherdata"]["npc"]):
                    for s in data["otherdata"]["slabels"]:
                        r.write(
                            "conc('"
                            + str(i + 1)
                            + "','"
                            + s
                            + "')="
                            + str(data["concdata"][s][i])
                            + ";\n"
                        )
                        r.write(
                            "rate('"
                            + str(i + 1)
                            + "','"
                            + s
                            + "')="
                            + str(data["ratedata"][s][i])
                            + ";\n"
                        )

        elif "time" in data["procdata"]:
            r.write("parameters conc(time,s),rate(time,s),aterm(time,m);\n")
            for j in range(ndatadic["time"]):
                for s in data["otherdata"]["slabels"]:
                    r.write(
                        "conc('"
                        + str(j + 1)
                        + "','"
                        + s
                        + "')="
                        + str(data["concdata"][s][0][j])
                        + ";\n"
                    )
                    r.write(
                        "rate('"
                        + str(j + 1)
                        + "','"
                        + s
                        + "')="
                        + str(data["ratedata"][s][0][j])
                        + ";\n"
                    )
                for m in range(len(data["otherdata"]["clc"])):
                    r.write(
                        "aterm('"
                        + str(j + 1)
                        + "','"
                        + str(m + 1)
                        + "')="
                        + str(aterm[0, j, m])
                        + ";\n"
                    )
        if "ads" in data["otherdata"]["mechanisms"]:
            for i in data["otherdata"]["adslabel"]:
                r.write("parameter " + i + "(pc,s);\n")
                for j in data["otherdata"]["slabels"]:
                    r.write(
                        i
                        + "(pc,'"
                        + j
                        + "')="
                        + data["procdata"]["adata"][i][j]
                        + ";\n"
                    )
        if 1 == 0:
            #        if ( 'adata' in data['procdata'] ):
            r.write("set adss(s) / ")
            for s in data["otherdata"]["adsspec"]:
                r.write(s)
                if s != data["otherdata"]["adsspec"][-1]:
                    r.write(",")
            r.write(" /;\n")
            r.write("set noads(s) / ")
            for s in [
                val
                for val in data["otherdata"]["slabels"]
                if val not in data["otherdata"]["adsspec"]
            ]:
                r.write(s)
                if (
                    s
                    != [
                        val
                        for val in data["otherdata"]["slabels"]
                        if val not in data["otherdata"]["adsspec"]
                    ][-1]
                ):
                    r.write(",")
            r.write(" /;\n")
            r.write("equations adeq, adyeq;\n")
            # This section write kads(pc,rx)=
            for i in data["otherdata"]["adslabel"]:
                r.write("parameter " + i + "(pc,s);\n")
                for j in data["otherdata"]["slabels"]:
                    r.write(
                        i
                        + "(pc,'"
                        + j
                        + "')="
                        + data["procdata"]["adata"][i][j]
                        + ";\n"
                    )

            if "time" in data["procdata"]:
                r.write("parameter aax(pc,rx,time);\n")
                for n in range(data["otherdata"]["npc"]):
                    for x in range(data["rxnet"]["nrx"]):
                        for t in range(data["otherdata"]["ntimedata"]):
                            for s in data["otherdata"]["slabels"]:
                                if (
                                    data["rxnet"][x, s] < 0
                                    and s in data["otherdata"]["adsspec"]
                                ):
                                    r.write(
                                        "aax('"
                                        + str(n + 1)
                                        + "','"
                                        + str(x + 1)
                                        + "','"
                                        + str(t + 1)
                                        + "')=kads('"
                                    )
                                    r.write(
                                        str(n + 1)
                                        + "','"
                                        + s
                                        + "')/(1+sum(s1,conc('"
                                        + str(n + 1)
                                        + "','"
                                        + str(t + 1)
                                        + "',s1)"
                                    )
                                    r.write("*kads('" + str(n + 1) + "',s1)));\n")
                                elif (
                                    data["rxnet"][x, s] < 0
                                    and s not in data["otherdata"]["adsspec"]
                                ):
                                    #  else:
                                    r.write(
                                        "aax('"
                                        + str(n + 1)
                                        + "','"
                                        + str(x + 1)
                                        + "','"
                                        + str(t + 1)
                                        + "')=1.0;\n"
                                    )
                r.write("variable adsaux(pc,rx,time);\n")
                r.write("adsaux.l(pc,rx,time)=1;\n")
            else:
                r.write("parameter aax(pc,rx);\n")
                for n in range(data["otherdata"]["npc"]):
                    for x in range(data["rxnet"]["nrx"]):
                        swtch = 0
                        for s in data["otherdata"]["slabels"]:
                            if (
                                (data["rxnet"][x, s] < 0)
                                and s in data["otherdata"]["adsspec"]
                                and swtch == 0
                            ):
                                r.write(
                                    "aax('"
                                    + str(n + 1)
                                    + "','"
                                    + str(x + 1)
                                    + "')=kads('"
                                )
                                r.write(
                                    str(n + 1)
                                    + "','"
                                    + s
                                    + "')/(1+sum(s1,conc('"
                                    + str(n + 1)
                                    + "',s1)"
                                )
                                r.write("*kads('" + str(n + 1) + "',s1)));\n")
                                swtch = 1
                            elif (
                                (data["rxnet"][x, s] < 0)
                                and s not in data["otherdata"]["adsspec"]
                                and swtch == 0
                            ):
                                r.write(
                                    "aax('"
                                    + str(n + 1)
                                    + "','"
                                    + str(x + 1)
                                    + "')=1.0;\n"
                                )
                                swtch = 1
                r.write("variable adsaux(pc,rx);\n")
                r.write("adsaux.l(pc,rx)=1;\n")

        # Get that equilibrium info in there too
        temp1 = 0
        if "revdata" in data["procdata"]:
            for i in data["otherdata"]["revlabel"]:
                jj = 0
                r.write("parameter " + i + "(pc,rx);\n")
                for rx in range(data["rxnet"]["nrx"]):
                    if rx in data["rxnet"]["rev"]:
                        j = data["otherdata"]["revspec"][jj]
                        r.write(
                            i
                            + "(pc,'"
                            + str(int(rx) + 1)
                            + "')="
                            + data["procdata"]["revdata"][i][j]
                            + ";\n"
                        )
                        jj = jj + 1
                    else:
                        r.write(i + "(pc,'" + str(int(rx) + 1) + "')=" + str(1) + ";\n")

        if 1 == 0:
            #        if ('revdata' in data['procdata']):
            temp1 = "set rvx1(rx) / 1"
            temp2 = "set rvx2(rx) / 2"
            for i in range(3, data["rxnet"]["nrx"]):
                if i % 2 == 0:
                    temp2 = temp2 + ", " + str(i)
                else:
                    temp1 = temp1 + ", " + str(i)
            temp1 = temp1 + " /;\n"
            temp2 = temp2 + " /;\n"
            r.write(temp1)
            r.write(temp2)

        r.write("parameter sigest(s);\n")
        sc = 0
        for s in data["otherdata"]["slabels"]:
            try:
                tempn = str(data["otherdata"]["noise"][sc])
            except Exception:
                tempn = "0"
            r.write("sigest('" + s + "')=" + tempn + ";\n")
            sc = sc + 1

        r.write("parameter rmat(m,rx,s),pmat(m,rx,s),revp(m,rx,s);\n")
        r.write("revp(m,rx,s) = 0;\n")
        for i in range(data["rxnet"]["nrx"]):
            for j in range(data["rxnet"]["mechanisms"]):
                if "clc" in data["otherdata"]:
                    mech = data["otherdata"]["clc"][j]
                else:
                    mech = data["otherdata"]["mechanisms"][j]
                for s in data["otherdata"]["slabels"]:
                    if 1 == 1:
                        # first handle reversible reactions - we are going to use the pmat for ma the reactants
                        if mech == "rev":
                            if data["rxnet"][i, s] < 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                            elif data["rxnet"][i, s] > 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                            else:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                        elif mech == "ma":
                            if data["rxnet"][i, s] < 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(abs(data["rxnet"][i, s]))
                                    + ";\n"
                                )
                            elif data["rxnet"][i, s] > 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                            else:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                        elif mech == "homcat":
                            r.write(
                                "rmat('"
                                + str(j + 1)
                                + "','"
                                + str(i + 1)
                                + "','"
                                + s
                                + "')="
                                + str(data["rxnet"][i, s])
                                + ";\n"
                            )
                            r.write(
                                "pmat('"
                                + str(j + 1)
                                + "','"
                                + str(i + 1)
                                + "','"
                                + s
                                + "')="
                                + str(data["rxnet"][i, s])
                                + ";\n"
                            )
                        elif mech == "ads":
                            if data["rxnet"][i, s] < 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(abs(data["rxnet"][i, s]))
                                    + ";\n"
                                )
                            elif data["rxnet"][i, s] > 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                            else:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                        elif mech == "iT":
                            if data["rxnet"][i, s] < 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(abs(data["rxnet"][i, s]))
                                    + ";\n"
                                )
                            elif data["rxnet"][i, s] > 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                            else:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                        elif "clc" in data["otherdata"]:
                            #                            for q in data['otherdata']['clc']:
                            if data["rxnet"][i, s] < 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(abs(data["rxnet"][i, s]))
                                    + ";\n"
                                )
                            elif data["rxnet"][i, s] > 0:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(data["rxnet"][i, s])
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                            else:
                                r.write(
                                    "rmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                                r.write(
                                    "pmat('"
                                    + str(j + 1)
                                    + "','"
                                    + str(i + 1)
                                    + "','"
                                    + s
                                    + "')="
                                    + str(0)
                                    + ";\n"
                                )
                                # j=j+1
                    else:
                        r.write(
                            "rmat('"
                            + str(j + 1)
                            + "','"
                            + str(i + 1)
                            + "','"
                            + s
                            + "')=0;\n"
                        )
                        r.write(
                            "pmat('"
                            + str(j + 1)
                            + "','"
                            + str(i + 1)
                            + "','"
                            + s
                            + "')=0;\n"
                        )

        r.write("binary variables y(rx,m);\n")
        if 1 == 0:
            #        if ( 'adata' in data['procdata']):
            r.write("binary variable yads;\n")
        r.write("positive variables k(rx,m);\n")
        if data["otherdata"]["npc"] > 0:
            if "T" in data["otherdata"]["pclabels"]:
                r.write("positive variables E(rx,m);\n")
        r.write("variable modelcomp;\n")
        if "time" in data["procdata"]:
            r.write("variables estrate(pc,time,s);\n")
            r.write("equations er(pc,time,s);\n")
        else:
            r.write("variables estrate(pc,s);\n")
            r.write("equations er(pc,s);\n")

        if data["otherdata"]["eobj"] == 1:
            if "time" in data["procdata"]:
                r.write("positive variables err(pc,time,s);\n")
            else:
                r.write("positive variables err(pc,s);\n")
            r.write("equations res1,res2;\n")

        if "time" in data["procdata"]:
            if 1 == data["otherdata"]["eobj"]:
                r.write(
                    "obj.. modelcomp =e= sum(pc,sum(s,sum(time,err(pc,time,s))));\n"
                )
            else:
                r.write(
                    "obj.. modelcomp =e= sum(pc,sum(s,sum(time,power(rate(pc,time,s)-estrate(pc,time,s),2))));\n"
                )
            # first write irrev model
            if "clc" in data["otherdata"]:
                if 1 == data["otherdata"]["eobj"]:
                    r.write(
                        "  res1(pc,time,s).. rate(pc,time,s) - estrate(pc,time,s) =l= err(pc,time,s);\n"
                    )
                    r.write(
                        "  res2(pc,time,s).. estrate(pc,time,s) - rate(pc,time,s) =l= err(pc,time,s);\n"
                    )
                r.write("  er(pc,time,s).. estrate(pc,time,s) =e=\n")
                if data["otherdata"]["npc"] > 1:
                    r.write(
                        "  sum(mclc,(sum(rx,(rmat(mclc,rx,s)*k(rx,mclc)*exp((-1*E(rx,mclc))/(8.314*(temp(pc))))*\n"
                    )
                else:
                    r.write("  sum(mclc,(sum(rx,(rmat(mclc,rx,s)*k(rx,mclc)*\n")
                r.write("  aterm(pc,time,mclc)))));\n")
            else:
                # Need to write independent T
                r.write("  er(pc,time,s).. estrate(pc,time,s) =e=\n")
                r.write(
                    "  (sum(rx,(rmat('1',rx,s)*k(rx,'1')*exp(-E(rx,'1')/(8.314*temp(pc)))*\n"
                )
                r.write("  aterm(pc,time,'1'))));\n")
            # Write the adsorption parameter is needed
            if 1 == 0:
                #            if ( 'adata' in data['procdata']):
                r.write("adsaux(pc,rx,time)*\n")
            if "rev" in data["otherdata"]["mechanisms"]:
                r.write("rmat('2',rx,s)*k(rx,'2')*exp(-E(rx,'2')/(8.314*temp(pc)))*\n")
                if "adata" in data["procdata"]:
                    r.write("adsaux(pc,rx,time)*\n")
                r.write("  (prod(s1,power(conc(pc,time,s1),pmat('2',rx,s1)))\n")
                r.write(
                    "-(prod(s3,power(conc(pc,time,s3),revp('2',rx,s3)))/keq(pc))))))),2))));\n"
                )
        else:
            mcount = 1
            # Input reactor is not dynamic
            r.write(
                "obj.. modelcomp =e= sum(pc,sum(s,(1/sigest(s))*power((rate(pc,s)-estrate(pc,s)),2)));\n"
            )
            # Estiamted rate
            r.write("  er(pc,s).. estrate(pc,s) =e=\n")
            if "iT" in data["otherdata"]["mechanisms"]:
                mcount = 1
                r.write(
                    "  sum(rx,(rmat('"
                    + str(mcount)
                    + "',rx,s)*k(rx,'"
                    + str(mcount)
                    + "')*\n"
                )
                if len(data["otherdata"]["mechanisms"]) > 1:
                    r.write(
                        "  (prod(s1,power(conc(pc,s1),pmat('"
                        + str(mcount)
                        + "',rx,s1)))))+(\n"
                    )
                else:
                    r.write(
                        "  (prod(s1,power(conc(pc,s1),pmat('"
                        + str(mcount)
                        + "',rx,s1))))));\n"
                    )
            if "ma" in data["otherdata"]["mechanisms"]:  # first write irrev models
                r.write(
                    "  sum(rx,(rmat('1',rx,s)*k(rx,'1')*exp(-E(rx,'1')/(8.314*temp(pc)))*\n"
                )
                r.write("  (prod(s1,power(conc(pc,s1),pmat('1',rx,s1)))))+(\n")
            #            if ( 1==0):
            #            if ( 'adata' in data['procdata']):
            #                r.write("adsaux(pc,rx)*\n")

            # now rev rxns
            if "rev" in data["otherdata"]["mechanisms"]:
                r.write("rmat('2',rx,s)*k(rx,'2')*exp(-E(rx,'2')/(8.314*temp(pc)))*\n")
                if 1 == 0:
                    #                if ( 'adata' in data['procdata']):
                    r.write("adsaux(pc,rx)*\n")
                r.write("  (prod(s1,power(conc(pc,s1),pmat('1',rx,s1)))\n")
                r.write(
                    "-(prod(s3,power(conc(pc,s3),pmat('2',rx,s3)))/keq(pc,rx))))+(\n"
                )
            # r.write("-(prod(s3,power(power(conc(pc,s3),revp('2',rx,s3)),pmat('2',rx,s3))/keq(pc,rx)))))+(\n")
            # Now the homcat models
            if "homcat" in data["otherdata"]["mechanisms"]:
                r.write("rmat('3',rx,s)*k(rx,'3')*exp(-E(rx,'3')/(8.314*temp(pc)))*\n")
                if 1 == 0:
                    # if ( 'adata' in data['procdata']):
                    r.write("adsaux(pc,rx)*\n")
                r.write("  (prod(s1,power(conc(pc,s1),pmat('3',rx,s1)))))+(\n")
            #            r.write("-(prod(s3,power(power(conc(pc,s3),revp('3',rx,s3)),pmat('2',rx,s3))/keq(pc,rx)))))))),2)));\n")
            # Now surface reactions
            if "ads" in data["otherdata"]["mechanisms"]:
                r.write("rmat('4',rx,s)*k(rx,'4')*exp(-E(rx,'4')/(8.314*temp(pc)))*\n")
                r.write(
                    "sum(s1,kads(pc,s1)*conc(pc,s1))/power(1+sum(s1,kads(pc,s1)*conc(pc,s1)),2)*\n"
                )
                r.write("  (prod(s1,power(conc(pc,s1),pmat('1',rx,s1))))));\n")
        #            r.write("))),2)));\n")

        r.write("cardcon.. sum(rx,sum(m,y(rx,m)))=e=" + str(ccon) + ";\n")
        # Set variable bounds here !!
        # Big M constraints big-M
        r.write("m1l(rx,m).. k(rx,m)=g=0*y(rx,m);\n")
        r.write("m1u(rx,m).. k(rx,m)=l=20*y(rx,m);\n")
        if data["otherdata"]["npc"] > 1:
            if "T" in data["otherdata"]["pclabels"]:
                r.write("m2l(rx,m).. E(rx,m)=g=500*y(rx,m);\n")
                r.write("m2u(rx,m).. E(rx,m)=l=10000*y(rx,m);\n")
        r.write("mc(rx).. sum(m,y(rx,m))=l=1;\n")
        # Disqualify irrev back rx if rev is chosen
        if "rev" in data["otherdata"]["mechanisms"]:
            for i in range(data["rxnet"]["nrx"]):
                if i in data["rxnet"]["rev"]:
                    r.write("equation req" + str(i + 1) + ";\n")
                    r.write(
                        "req"
                        + str(i + 1)
                        + ".. y('"
                        + str(i + 1)
                        + "','2')+y('"
                        + str(i + 2)
                        + "','1')=l=1;\n"
                    )
        # handle adsorption binary
        if 1 == 0:
            #        if ('adata' in data['procdata']):
            if "time" not in data["procdata"]:
                r.write("adeq(pc,rx).. adsaux(pc,rx) =e=")
                r.write(" yads*aax(pc,rx)+(1-yads)*1;\n")
                r.write("adyeq.. yads =l= 1;\n")
            else:
                r.write("adeq(pc,rx,time).. adsaux(pc,rx,time) =e=")
                r.write(" yads*aax(pc,rx,time)+(1-yads)*1;\n")
                r.write("adyeq.. yads =l= 1;\n")

        r.write("model ccmodel /all/;\n")
        # fix bad binaries
        #        if('clc' not in data['otherdata']):
        #            for i in range(data['rxnet']['nrx']):
        #                for j in range(len(data['rxnet']['mechanisms'])):
        #                    if ( i not in data['rxnet'][data['otherdata']['mechanisms'][j]]):
        #                        r.write("y.fx('"+str(i+1)+"','"+str(j+1)+"')=0;\n")

        #        r.write("option minlp=dicopt, reslim=1000,optca=1e-7,optcr=1e-7;\n")
        if 0:
            r.write("y.l('1','19')=1;\n")
            r.write("y.fx('1','20')=0;\n")
            r.write("y.fx('1','21')=0;\n")
            r.write("y.fx('1','18')=0;\n")
            r.write("y.fx('1','17')=0;\n")
            r.write("y.fx('1','16')=0;\n")
            r.write("y.fx('1','15')=0;\n")
        #            r.write("k.l('1','19')="+str(10926)+";\n")
        #            r.write("E.l('1','19')="+str(55319)+";\n")
        elif debug["true"]["fix"]:
            for i in range(1, len(debug["true"]["k"]) + 1):
                r.write("y.fx('" + str(i) + "','1')=1;\n")
                r.write(
                    "k.fx('"
                    + str(i)
                    + "','1')="
                    + str(debug["true"]["k"][i - 1])
                    + ";\n"
                )
                if data["otherdata"]["npc"] > 1:
                    r.write(
                        "E.fx('"
                        + str(i)
                        + "','1')="
                        + str(debug["true"]["e"][i - 1])
                        + ";\n"
                    )
        #    r.write("solve ccmodel minimizing modelcomp using minlp;\n")
        r.write("option minlp=baron, reslim=10000,optca=1e-7,optcr=1e-7;\n")
        r.write("solve ccmodel minimizing modelcomp using minlp;\n")
        if "time" in data["procdata"]:
            r.write("display estrate.l, rate;\n")
        else:
            r.write("display rate;\n")
        if "clc" in data["otherdata"]:
            # rates.put is estrate then rate_obs
            r.write("file rates;\n")
            r.write("put rates;\n")
            r.write("  loop(pc,loop(time,\n")
            r.write("  put ord(pc),ord(time),estrate.l(pc,time,'X'):20:10\n")
            r.write("  , rate(pc,time,'X'):20:10\n")
            r.write("  , rate(pc,time,'X'):20:10\n")
            r.write("/););\n")
        elif "T" not in data["otherdata"]["pclabels"]:
            r.write("parameter resid(s);\n")
            r.write(
                "resid(s) = sum(pc, power( rate(pc,s) - (sum(rx,(rmat('1',rx,s)*k.l(rx,'1')*(prod(s1,power(conc(pc,s1),pmat('1',rx,s1))))))),2));\n"
            )
            r.write("file rates;\n")
            r.write("put rates;\n")
            r.write("  loop(s,\n")
            r.write("  put resid(s):20:10);\n")
            r.write(" putclose rates;\n")
        else:
            r.write("parameter resid(s);\n")
            r.write("resid(s) = sum(pc, power( rate(pc,s) - estrate.l(pc,s),2));\n")
            r.write("file rates;\n")
            r.write("put rates;\n")
            r.write("  loop(s,\n")
            r.write("  put resid(s):20:10);\n")
            r.write(" putclose rates;\n")

        r.write("file res;\n")
        r.write("put res;\n")
        r.write("res.pc=4;\n")
        r.write("put modelcomp.l:20:10 / ;\n")
        r.write("  loop(rx,loop(m,\n")
        r.write("  put ord(rx),ord(m),y.l(rx,m):1:0\n")
        r.write("  , k.l(rx,m):20:10")
        if "T" in data["otherdata"]["pclabels"]:
            r.write(",E.l(rx,m):20:10")
        r.write("\n")
        r.write("/););\n")
        if 0:
            #        if( 'adata' in data['procdata']):
            if "time" in data["procdata"]:
                r.write("  loop(pc,loop(rx,loop(time,\n")
                r.write("  put ord(rx), adsaux.l(pc,rx,time):20:10/);););\n")
            else:
                r.write("  loop(pc,loop(rx,\n")
                r.write("  put ord(rx), adsaux.l(pc,rx):20:10/););\n")
        r.write("putclose res;\n")

    return aterm



import os
import logging
import tarfile
import idaes
from shutil import copyfile
from pyomo.common.download import FileDownloader

_log = logging.getLogger(__name__)

def download_binaries(url=None, verbose=False):
    """
    Download IDAES solvers and libraries and put them in the right location. Need
    to supply either local or url argument.

    Args:
        url (str): a url to download binary files to install files

    Returns:
        None
    """
    if verbose:
        _log.setLevel(logging.DEBUG)
    idaes._create_lib_dir()
    idaes._create_bin_dir()
    solvers_tar = os.path.join(idaes.bin_directory, "idaes-solvers.tar.gz")
    libs_tar = os.path.join(idaes.lib_directory, "idaes-lib.tar.gz")
    fd = FileDownloader()
    arch = fd.get_sysinfo()
    if url is not None:
        if not url.endswith("/"):
            c = "/"
        else:
            c = ""
        solvers_from = c.join([url, "idaes-solvers-{}-{}.tar.gz".format(arch[0], arch[1])])
        libs_from = c.join([url, "idaes-lib-{}-{}.tar.gz".format(arch[0], arch[1])])
        _log.debug("URLs \n  {}\n  {}\n  {}".format(url, solvers_from, libs_from))
        _log.debug("Destinations \n  {}\n  {}".format(solvers_tar, libs_tar))
        if arch[0] == 'darwin':
            raise Exception('Mac OSX currently unsupported')
        fd.set_destination_filename(solvers_tar)
        fd.get_binary_file(solvers_from)
        fd.set_destination_filename(libs_tar)
        fd.get_binary_file(libs_from)
    else:
        raise Exception("Must provide a location to download binaries")

    _log.debug("Extracting files in {}".format(idaes.bin_directory))
    with tarfile.open(solvers_tar, 'r') as f:
        f.extractall(idaes.bin_directory)
    _log.debug("Extracting files in {}".format(idaes.lib_directory))
    with tarfile.open(libs_tar, 'r') as f:
        f.extractall(idaes.lib_directory)



import logging

def getIdaesLogger(name, level=None):
    """ Return an idaes logger.

    Args:
        name: usually __name__
        level: standard IDAES logging level (default use IDAES config)

    Returns:
        logger
    """
    # this function is fairly useless right now, but it helps to standardize

    # This probably does nothing, but if you ask for an idaes logger outside
    # the idaes package it makes sure you get one.
    name = ".".join(["idaes", name.lstrip("idaes.")])
    l = logging.getLogger(name)
    if level is not None:
        l.setLevel(level)
    return logging.getLogger(name)

def solver_tee(logger, tee_level=logging.DEBUG):
    """Function to produce solver output based on the logging level of a specific
    logger. This function just helps standardize the level for solver output to
    appear and make code a bit cleaner.

    Args:
        logger: logger to get output level from
        tee_level: Level at which to show solver output, usually use default

    Returns
        (bool)
    """
    return logger.getEffectiveLevel() <= tee_level

def init_tee(logger, tee_level=2):
    """Function to use in initialization to determine at a given output level
    whether to use the sovler tee option to print solver output. This function
    just helps standardize the level for solver output to appear and make the
    initialization routine code a bit cleaner.

    Args:
        logger: logger to get output level from
        tee_level: Level at which to show solver output, usually use default

    Returns
        (bool)
    """
    return logger.getEffectiveLevel() <= tee_level

def condition(res):
    """Get the solver termination condition.  Since it seems to be common to
    have an if block to check for None if the solver call raised a handled
    exception"""

    if res is None:
        return "Error, no result"
    elif isinstance(res, str):
        return res
    else:
        return str(res.solver.termination_condition)

def getInitLogger(name, level=None):
    """ Get a model initialization logger

    Args:
        name: Object name (usually Pyomo Component name)
        level: Logging detail level (for initialization routines 1 to 6)
             * 0 = Use default idaes.init logger setting
             * 1 = Maximum output
             * 2 = Include solver output
             * 3 = Return solver state for each step in subroutines
             * 4 = Return solver state for each step in routine
             * 5 = Final initialization status and exceptions
             * 6 = No output

    Returns:
        logger
    """
    name = ".".join(["idaes.init", name])
    l = logging.getLogger(name)
    if level is not None:
        l.setLevel(level)
    return l

def getModelLogger(name, level=None):
    """ Get a logger for an IDAES model. This function helps users keep their
    loggers in a standard location and using the IDAES logging config.

    Args:
        name: Name (usually __name__).  Any starting 'idaes.' is stripped off, so
            if a model is part of the idaes package, idaes won't be repeated.
        level: Standard Python logging level (default use IDAES config)

    Returns:
        logger
    """
    name = ".".join(["idaes.model", name.lstrip("idaes.")])
    l = logging.getLogger(name)
    if level is not None:
        l.setLevel(level)
    return logging.getLogger(name)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for DMF 'workspace' module.
"""
# stdlib
import logging
import os
import sys

# third-party
import pytest

# local
from idaes.dmf import workspace, errors
from idaes.dmf.util import TempDir

# for testing
from .util import init_logging

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)

BADPATH = os.path.join("should", "never", "exist", "seriously", "no")


@pytest.fixture
def wsdir():
    with TempDir() as d:
        yield d
        print("Removing temporary workspace directory: {}".format(d))


# Tests
# -----


def test_ws_init_notfound():
    with pytest.raises(errors.WorkspaceNotFoundError):
        workspace.Workspace(BADPATH)


def test_ws_init_noconf(wsdir):
    with pytest.raises(errors.WorkspaceConfNotFoundError):
        workspace.Workspace(wsdir)


def test_ws_init_badconf(wsdir):
    conf = open(os.path.join(wsdir, workspace.Workspace.WORKSPACE_CONFIG), "w")
    conf.write("note: this config is wack\n")
    conf.close()
    with pytest.raises(errors.WorkspaceConfMissingField):
        workspace.Workspace(wsdir)


def test_ws_init_create_nodefaults(wsdir):
    ws = workspace.Workspace(wsdir, create=True)
    # if defaults were set, then HTML documentation paths
    # will not be empty, so test this value
    assert ws.get_doc_paths() == []


def test_ws_init_create_defaults(wsdir):
    ws = workspace.Workspace(wsdir, create=True, add_defaults=True)
    # if defaults were set, then HTML documentation paths
    # will not be empty, so test this value
    assert ws.get_doc_paths() != []


def test_ws_accessors(wsdir):
    ws = workspace.Workspace(wsdir, create=True, add_defaults=True)
    assert ws.wsid != ""
    assert os.path.normpath(ws.root) == os.path.normpath(wsdir)
    _ = ws.name
    _ = ws.description




##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# This file contains two subroutines
#  - construcutmodel(results, kwargs) : constructs a steady-state pyomo reactor model for use in error maximization sampling
#  - ems(riperes,sim,lb,ub,nspec,kwargs) : perform error maximization sampling
import numpy as np
import rbfopt
import os
# import sys
# import random
import datetime
import time
# pkg
from idaes.surrogate import ripe

def constructmodel(riperes, **kwargs):
    # Inputs:
    # riperes - output of ripemodel()
    # kwargs  - kwargs supplied to ripemodel()
    # Outputs:
    # ripesim - pyomo model

    # sharedata must be supplied explicitly
    if "sharedata" in kwargs.keys():
        sharedata = kwargs["sharedata"]
    else:
        sharedata = ripe.sharedata

    # Define gas constant for clarity later
    gc = sharedata["gasconst"]

    # determine type of kinetic form used
    if "E" in riperes.keys():
        acte = riperes["E"]
        if "Tref" in kwargs.keys():
            Tref = sharedata["Tref"]
            ptype = 3
        else:
            ptype = 2
    else:
        ptype = 1
    prek = riperes["k"]
    mechs = riperes["mechanisms"]
    stoichs = riperes["stoichiometry"]
    ns = len(stoichs[0])
    nr = len(mechs)
    # replace 'massact' with species specific mechanisms (required of rbfopt)
    for i in range(nr):
        if mechs[i] == "massact":
            mechs[i] = ripe.mechs.mechperstoich(mechs[i], stoichs[i])

    # Define a steady-state reactor model in pyomo using ripe resutls
    def ripesim(data):
        import pyomo.environ as pyo

        # ripesim expects input in a particular order
        model = pyo.ConcreteModel()
        # s is index over species
        model.s = pyo.RangeSet(ns)
        # r over reactions
        model.r = pyo.RangeSet(nr)
        # initialize variable from data
        model.conc0 = pyo.Param(
            model.s, initialize=dict(((s), data[s - 1]) for s in model.s)
        )
        if ptype > 1:
            model.T = pyo.Param(initialize=data[ns])
            model.E = pyo.Param(
                model.r, initialize=dict(((r), acte[r - 1]) for r in model.r)
            )

        model.nu = pyo.Param(
            model.r,
            model.s,
            initialize=dict(
                ((r, s), stoichs[r - 1][s - 1]) for r in model.r for s in model.s
            ),
        )
        model.conc = pyo.Var(model.s, domain=pyo.NonNegativeReals, initialize=1.0)
        try:
            model.flow = pyo.Param(
                model.s, initialize=dict(((s), data[ns + s]) for s in model.s)
            )
        except Exception:
            model.flow = pyo.Param(model.s, initialize=1.0)
        try:
            model.vol = pyo.Param(initialize=float(data[2 * ns + 1]))
        except Exception:
            model.vol = pyo.Param(initialize=1.0)
        model.k = pyo.Param(
            model.r, initialize=dict(((r), prek[r - 1]) for r in model.r)
        )
        model.rate = pyo.Var(model.r, initialize=0.0)
        model.dum = pyo.Var(model.s, initialize=0.0)

        # define predicted rates of generation
        # different problem types require different rate forms
        def prates_1(model, r):
            return model.rate[r] == model.k[r] * mechs[r - 1](*model.conc[:])

        def prates_2(model, r):
            indata = [model.conc[i] for i in model.s] + [data[ns]]
            return model.rate[r] == model.k[r] * pyo.exp(
                -(model.E[r] / (gc * model.T))
            ) * mechs[r - 1](*indata)

        def prates_3(model, r):
            indata = [model.conc[i] for i in model.s] + [data[ns]]
            return model.rate[r] == model.k[r] * pyo.exp(
                -(model.E[r] / (gc)) * (1 / (model.T) - 1 / (Tref))
            ) * mechs[r - 1](*indata)

        if ptype == 1:
            trule = prates_1
        elif ptype == 2:
            trule = prates_2
        else:
            trule = prates_3
        model.rc = pyo.Constraint(model.r, rule=trule)

        def balance(model, s):
            return 100.0 * model.dum[s] == model.flow[s] * (1.0 / model.vol) * (
                model.conc0[s] - model.conc[s]
            ) + sum([model.nu[i, s] * model.rate[i] for i in model.r])

        model.bal = pyo.Constraint(model.s, rule=balance)

        def obj(model):
            return sum([model.dum[i] ** 2 for i in model.s])

        model.OBJ = pyo.Objective(rule=obj)
        opt = pyo.SolverFactory(sharedata["minlp_path"])
        results = opt.solve(model, tee=sharedata["showpyomo"])
        model.solutions.store_to(results)
        conres = []
        for i in range(ns):
            conres.append(
                results.Solution.Variable["conc[" + str(i + 1) + "]"]["Value"]
            )
        return conres

    return ripesim


def ems(riperes, sim, lb, ub, nspec, **kwargs):
    # This subroutine performs error maximization sampling
    # Inputs:
    # riperes - results from ripemodel()
    # sim     - Original black-box simulator (callable in python)
    # lb/ub   - bounds for each independent variable
    # nspec   - number of species (can be different than #lb/ub)
    # Outputs:
    # x       - next best input point
    # errs    - absolute errors obtained on predicted point

    # Non-default options can be provided explicitly through kwargs
    if "sharedata" in kwargs.keys():
        sharedata = kwargs["sharedata"]
    else:
        sharedata = ripe.sharedata
    # poskeys = sharedata["ivars"] + ["x"]
    inkeys = list(set(kwargs.keys()) - set(["sharedata"]))
    ndim = len(lb)
    ns = nspec

    # Call atermconstruct.formatinputs to get input data
    # in a convenient form
    if "x" in inkeys:
        conc = kwargs["x"]
        dflag = True
    else:
        conc = [1] * ns
        dflag = False

    if "frac" in kwargs.keys():
        dofrac = True
        # conc[i] = kwargs["frac"](conc[i])  # i not defined?
    else:
        dofrac = False
    data, kwargs, fdata, pc, alldata = ripe.atermconstruct.formatinputs(conc, kwargs)

    # Handle inkeys exceptions and (possibly) generate a pyomo model
    if "Tref" in inkeys:
        Tref = sharedata["Tref"]
    if "res_sim" in inkeys:
        res_sim = kwargs["res_sim"]
    else:
        if "Tref" in inkeys:
            res_sim = constructmodel(riperes, sharedata=sharedata, Tref=Tref)
        else:
            res_sim = constructmodel(riperes, sharedata=sharedata)

    ndim = len(lb)
    nd, ns = np.shape(fdata)

    if "T" in inkeys:
        params = [[], []]
        params[0] = riperes["k"]
        params[1] = riperes["E"]
    else:
        params = riperes["k"]

    # Check for multiple requested points
    # if "nreq" in inkeys:
    #     nreq = kwargs["nreq"]
    # else:
    #     nreq = 1

    # subroutine for using fractional arguments (mole fracs or partial pressure)
    def apply_frac(x, doit=True):
        if doit:
            return kwargs["frac"](x[:])
        else:
            return x

    def check_fun(x):
        # Check fun is used to evaluate errors of the current model on provided data
        x[:] = apply_frac(x[:], dofrac)
        return -1.0 * np.sum(
            np.divide(
                np.power(np.subtract(x[:ns], res_sim(x[ns:])), 2), riperes["sigma"]
            )
        )

    def max_err(x):
        # max_err determines the absolute errors (max called later)
        x[:] = apply_frac(x[:], dofrac)
        errs = np.absolute(np.subtract(sim(x), res_sim(x)))[0]
        return errs

    def error_fun(x):
        # error fun is the black box objective
        x[:] = apply_frac(x[:], dofrac)
        sim_conc = sim(x[:])
        res_conc = res_sim(x[:])
        return -1.0 * np.sum(
            np.power(np.divide(np.subtract(sim_conc, res_conc), sim_conc), 2)
        )

    t_targets = np.zeros([nd, 1])
    if dflag:
        for i in range(nd):
            t_targets[i] = check_fun(alldata[i][:])

    bb = rbfopt.RbfoptUserBlackBox(
        ndim, np.array(lb), np.array(ub), np.array(["R"] * ndim), error_fun
    )
    # Request that any point returned be no closer than the other closest point
    distance = 10.0
    for i in range(nd):
        for j in range(nd):
            if i != j:
                distance = max(
                    distance,
                    np.linalg.norm(
                        alldata[i, ns : ns + ndim] - alldata[j, ns : ns + ndim]
                    ),
                )
    settings = []
    t = datetime.datetime.now()
    d_mult = 1.0
    settings = rbfopt.RbfoptSettings(
        nlp_solver_path=sharedata["nlp_path"],
        minlp_solver_path=sharedata["minlp_path"],
        print_solver_output=False,
        min_dist=d_mult * distance,
        algorithm="Gutmann",
        rand_seed=int(time.mktime(t.timetuple())),
    )  # random seed required for RBFopt functionality
    # Alternative settings used in testing saved for posterity
    #    settings = rbfopt.RbfoptSettings( do_infstep=True,nlp_solver_path='/usr/local/ipopt/3.10.2/ipopt/ipopt.pc', minlp_solver_path='~/baron/baron',print_solver_output=False, algorithm='MSRM',global_search_method = 'solver',rand_seed = int(time.mktime(t.timetuple())), min_dist = d_mult*distance)
    # ),#init_strategy='all_corners',
    # num_global_searches = 0,)
    # provide initialization data if flag is tripped
    if not dflag:
        #        print 'check conc : ',alldata[:,ns:ns+ndim],t_targets
        alg = rbfopt.RbfoptAlgorithm(
            settings,
            bb,
            do_init_strategy=False,
            init_node_pos=alldata[:, ns : ns + ndim],
            init_node_val=t_targets,
        )  # ,num_nodes_at_restart=nd)
    else:
        alg = rbfopt.RbfoptAlgorithm(settings, bb)

    # Call to rbfopt in a manner that terminal is not polluted
    f = open(os.devnull, "w")
    alg.set_output_stream(f)
    val, new_x, itercount, evalcount, fast_evalcount = alg.optimize(pause_after_iters=1)
    f.close()
    if len(new_x) < ndim:
        x = list(new_x) + [0] * (ndim - len(new_x))
    else:
        x = list(new_x)

    # Calculate relevant metrics and identify output responsible for error violation
    errs = max_err(x)
    # maxv = np.max(errs)
    # loc = np.argmax(errs)

    if "T" in inkeys:
        # prop_x = x[:-1]
        prop_t = x[-1]
    # else:
    #     prop_x = x
    #    print 'Maximum error on new proposed point : ', maxv,' on species # ',loc+1
    #    print 'Proposed initial concentrations', prop_x
    if "T" in inkeys:
        print("Proposed Temperature", prop_t)

    return [x, errs]



# coding: utf-8
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

# stdlib
import importlib
import pathlib
import os
import re
import time

# package
import idaes
from idaes.dmf.util import ColorTerm

good_modname = re.compile(r"[a-zA-Z][a-zA-Z0-9_]*")


SKIP, OK, BAD = "skipped", "success", "failed"


_term = ColorTerm()


def print_path_status(path, status, msg=""):
    color = {SKIP: "", OK: _term.green, BAD: _term.red}[status]
    print(f"{color}{status.upper():8s}{_term.resetc} {path} {msg}")


def importr(root: pathlib.Path, max_sec=10):
    """Import r_ecursively from the given root path.
    """
    base, failures, total = root.parent, {}, 0
    # iterate over flattened list of all paths ending a Python source file
    for path in root.rglob("*.py"):
        # check that all path components are valid module names
        # - this is to skip directories like '.ipynb_checkpoints'
        bad = False
        for name in path.parts[:-1]:
            if name == os.path.sep:
                continue
            if (
                name.startswith("Workshop")       # skip workshop tutorials (1)
                or name.startswith("Module_")     # skip workshop tutorials (2)
                or name == "tests"                # skip all tests
                or not good_modname.match(name)   # no _bla
            ):
                bad = True

            if bad:
                break
        # stop if bad directory component or Python filename is invalid
        bad = bad or not good_modname.match(path.parts[-1][:-3])
        if bad:
            print_path_status(path, SKIP)
            continue
        # module is valid, try importing it now
        module_path = path.relative_to(base).with_suffix("")
        module_name = ".".join(module_path.parts)
        try:
            start = time.time()
            importlib.import_module(module_name)
            sec = time.time() - start
            if sec > max_sec:
                print_path_status(path, BAD, msg="time={sec:3.1f}s")
                raise ImportError(f"Import took too long ({sec:.1f}s)")
            print_path_status(path, OK)
        except ImportError as e:
            failures[module_name] = str(e)
            print_path_status(path, BAD, msg="import error")
        total += 1
    return failures, total


def test_import():
    root_dir = pathlib.Path(idaes.__file__).parent
    failures, total = importr(root_dir)
    n = len(failures)
    assert n == 0, f"{n:d} failures in {total:d} tests: {failures}"



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Standard IDAES Feed block with phase equilibrium.
"""
from enum import Enum

# Import Pyomo libraries
from pyomo.environ import Reference
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe

__author__ = "Andrew Lee"


# Enumerate options for material balances
class FlashType(Enum):
    isothermal = 1
    isenthalpic = 2


@declare_process_block_class("FeedFlash")
class FeedFlashData(UnitModelBlockData):
    """
    Standard Feed block with phase equilibrium
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Feed units do not support dynamic behavior."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Feed units do not have defined volume, thus this must be
False."""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("flash_type", ConfigValue(
        default=FlashType.isothermal,
        domain=In(FlashType),
        description="Type of flash to perform",
        doc="""Indicates what type of flash operation should be used.
**default** - FlashType.isothermal.
**Valid values:** {
**FlashType.isothermal** - specify temperature,
**FlashType.isenthalpic** - specify enthalpy.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))

    def build(self):
        """
        Begin building model.

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(FeedFlashData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(default={
                "dynamic": self.config.dynamic,
                "has_holdup": self.config.has_holdup,
                "property_package": self.config.property_package,
                "property_package_args": self.config.property_package_args})

        # No need for control volume geometry

        self.control_volume.add_state_blocks(
                has_phase_equilibrium=True)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=True)

        # Add isothermal constraint
        if self.config.flash_type == FlashType.isothermal:
            @self.Constraint(self.flowsheet().config.time,
                             doc="Isothermal constraint")
            def isothermal(b, t):
                return (b.control_volume.properties_in[t].temperature ==
                        b.control_volume.properties_out[t].temperature)
        elif self.config.flash_type == FlashType.isenthalpic:
            @self.Constraint(self.flowsheet().config.time,
                             doc="Isothermal constraint")
            def isenthalpic(b, t):
                cv = b.control_volume
                return (sum(cv.properties_in[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list) ==
                        sum(cv.properties_out[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list))

        self.control_volume.add_momentum_balances(
            balance_type=MomentumBalanceType.pressureTotal)

        # Add references to all feed state vars
        s_vars = self.control_volume.properties_in[
                    self.flowsheet().config.time.first()].define_state_vars()
        for s in s_vars:
            l_name = s_vars[s].local_name
            if s_vars[s].is_indexed():
                slicer = (
                    self.control_volume.properties_in[:].component(l_name)[...])
            else:
                slicer = self.control_volume.properties_in[:].component(l_name)

            r = Reference(slicer)
            setattr(self, s, r)

        # Add Ports
        self.add_outlet_port()

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
                {"Outlet": self.outlet},
                time_point=time_point)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import sys


def checkoptions(kwargs, sharedata, debug):
    # This subroutine checks kwargs for ripe-specific parameters
    # the subroutine checks for valid arguments, but only a subset
    # are check to initialize sharedata
    # Inputs:
    # kwargs     - kwargs from ripemodel()
    # sharedata  - dictionary shared across subroutines
    # debug      - Additional shared dictionary
    # Outputs:
    # sharedata  - Originally defined in shared.py
    #    this dictionary contains information pertaining to RIPE's function

    inkeys = list(kwargs)  # kwargs.keys() python 2
    for key in inkeys:
        if key not in sharedata["kwargsin"] + sharedata["ivars"]:
            sys.exit(
                "Keyword argument "
                + key
                + " is not recognized. Consult the documnetation for appropriate arguments"
            )
        elif key == "minlp_path" or key == "alamo_path":
            sharedata[key] = kwargs[key]
        elif key == "keepfiles":
            sharedata[key] = kwargs[key]
        elif key == "showpyomo":
            sharedata[key] = kwargs[key]
        elif key == "Tref" or key == "Tr" or key == "tr" or key == "tref":
            sharedata["Tref"] = kwargs[key]
            sharedata["Tr"] = kwargs[key]
        elif key == "return_model":
            sharedata[key] = kwargs[key]
        elif key == "mechanisms" or key == "mech" or key == "mechs":
            kwargs["mech"] = kwargs[key]
        elif key == "stoichiometry" or key == "stoich" or key == "stoichs":
            kwargs["stoich"] = kwargs[key]
        elif key == "hide_output":
            sharedata[key] = kwargs[key]
        elif key == "deltaterm":
            sharedata[key] = kwargs[key]
        elif key == "zscale":
            sharedata[key] = kwargs[key]
        elif key == "ascale":
            sharedata[key] = kwargs[key]
        elif key == "onemechper":
            sharedata[key] = kwargs[key]
        elif key == "expand_output":
            sharedata[key] = kwargs[key]
        elif key == "time" or key == "t":
            kwargs["t"] = kwargs[key]
        elif (
            key == "Temp"
            or key == "Temperature"
            or key == "temp"
            or key == "temperature"
        ):
            kwargs["T"] = kwargs[key]

    return sharedata, kwargs



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Base classes for visualization and plotting in IDAES.

Create new plots by inheriting from :class:`PlotBase`. See the
:mod:`idaes.vis.bokeh_plots` module for examples.
"""
# stdlib
from abc import ABC, abstractmethod
from typing import Type, List
# third party
import pandas as pd


class PlotBase(ABC):
    """Abstract base class for a plot.
    """

    def __init__(self, current_plot):
        """Constructor.

        Args:
            current_plot: Set the current plot.
        """
        self.current_plot = current_plot

    @abstractmethod
    def annotate(self, x, y, label: str):
        """Annotate a plot with a given point and a label.

        Args:
            x: Value of independent variable.
            y: Value of dependent variable.
            label: Text label.
        """
        pass

    @abstractmethod
    def resize(self, height: int = -1, width: int = -1):
        """Resize a plot's height and width.

        Args:
            height: Height in screen units.
            width: Width in screen units.

        Returns:
            None

        Raises:
            None
        """
        pass

    @abstractmethod
    def save(self, destination: str):
        """Save the current plot object to HTML in filepath provided by destination.

        Args:
            destination: Valid file path to save HTML to.

        Returns:
            filename where HTML is saved.

        Raises:
            None
        """
        pass

    @abstractmethod
    def show(self, in_notebook=True):
        """Display plot in a Jupyter notebook.

        Args:
            in_notebook: Display in Jupyter notebook or generate HTML file.

        Returns:
            None

        Raises:
            None
        """
        pass

    @classmethod
    def validate(cls, data_frame: pd.DataFrame, x: str, y: List, legend=None):
        """Validate that the plot parameters are valid.

        Args:
            data_frame: a pandas data frame of any type.
            x: Key in data-frame to use as x-axis.
            y: Keys in data-frame to use as y-axis.
            legend: List of labels to use as legend for a plot.
        Returns:
            True, '' on valid data frames (if x and y are in the data frame keys)
            False, "message" on invalid data
        """
        if data_frame is None or data_frame.empty:
            return False, "Empty data frame"
        if not legend and legend is not None:
            return False, "Bad legend labels"
        data_keys = set(data_frame.keys())
        if x not in data_keys:
            return False, f"The x-axis key {x} is not a column"
        if not set(y).issubset(data_keys):
            diff = set(y) - data_keys
            if len(diff) > 1:
                return False, f"The y-axis keys {diff} are not columns"
            else:
                return False, f"y-axis key {diff} is not a column"
        return True, ''


class PlotRegistry(object):
    """Set of associations between objects/classes + a plot name, and
    the parameters and values needed to perform the plot.

    The basic idea is to create a set of named plots associated with
    a given IDAES class, and then allow the user or other APIs to invoke that
    plot once the data is populated in an instance of the class. This keeps
    the details of how to create plots of a given type in the classes that
    will create them.

    For example::

        class MyIdaesClass(ProcessBase):
          # .. code for the class
          def plot_setup(self, plot_class):
              # .. details of creating plot_instance from object contents ..
              return plot_instance
        PlotRegistry().register(MyIdaesClass, 'basic', MyIdaesClass.plot_setup)

        # .. and, later ..
        obj = MyIdaesClass(...)
        # .. do things that fill "obj" with data ..
        # now create the plot
        plot = PlotRegistry().get(obj, 'basic')
        plot.show()

    XXX: This class is not actually used (yet) by any of the IDAES models.
    """

    __objs = {}  # shared state for all instances

    def __init__(self):
        self._objs = self.__objs

    def register(
        self, obj, name: str, plot: Type, setup_fn=None, overwrite: bool = False
    ):
        """Register an object/plot combination.

        Args:
            obj: Class or instance
            name: Name for this association
            plot: Plot class
            setup_fn: Optional setup function to call. Function should take
                      two arguments: `plot` class instance, `obj` assoc. with plot.
            overwrite: If true, allow overwrite of existing entry in the registry
        """
        if obj in self._objs:
            entry = self._objs[obj]
        else:
            entry = {}
            self._objs[obj] = entry
        if name in entry and not overwrite:
            raise KeyError(
                f"Cannot register {name}: would overwrite existing "
                f"entry for object, and 'overwrite' parameter is not set"
            )
        entry[name] = (plot, setup_fn)

    def get(self, obj, name: str):
        """Get a plot object for the given object + name.

        Args:
            obj: Object for which to get the plot
            name: Registered name of plot to get

        Returns:
            Return value of setup function given to :meth:`register()`,
            or, if that is empty, the retrieved plot object.
        """
        if obj not in self._objs:
            raise KeyError(f"No plots registered for {obj}")
        entry = self._objs[obj]
        if name not in entry:
            raise KeyError(f"No plot named '{name}' found in entry for object {obj}")
        plot, setup_fn = entry[name]
        if setup_fn is not None:
            result = setup_fn(plot, obj)
        else:
            result = plot
        return result

    def remove_all(self):
        """Remove all entries from the registry.

        Since the registry is a singleton, this removes all entries from ALL
        instances. Use with care.
        """
        for k in list(self._objs.keys()):
            del self._objs[k]



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
from idaes.core.property_base import PhysicalParameterBlock


class IndexMePlease1(PhysicalParameterBlock):

    @classmethod
    def define_metadata(cls, m):
        m.add_default_units({'temperature': 'K'})
        m.add_properties({'pressure': {'units': 'Pa', 'method': 'foo'},
                          'temperature': {'method': 'bar'}})



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Utility functions.
"""
# stdlib
import importlib
import json
from json import JSONDecodeError
import logging
import os
import re
import shutil
import tempfile
import time
import yaml

# third-party
import colorama

__author__ = "Dan Gunter"

_log = logging.getLogger(__name__)


def yaml_load(*args):
    return yaml.load(*args, Loader=yaml.SafeLoader)


def strlist(x, sep=", "):
    # type: (list, str) -> str
    return sep.join([str(item) for item in x])


def get_file(file_or_path, mode="r"):
    """Open a file for reading, or simply return the file object.
    """
    if hasattr(file_or_path, "read"):
        return file_or_path
    return open(file_or_path, mode)


def import_module(name):
    mod = importlib.import_module(name)
    return mod


def get_module_version(mod):
    """Find and return the module version.

    Version must look like a semantic version with
    `<a>.<b>.<c>` parts; there can be arbitrary extra
    stuff after the `<c>`. For example::

        1.0.12
        0.3.6
        1.2.3-alpha-rel0

    Args:
        mod (module): Python module
    Returns:
        (str) Version string or None if not found
    Raises:
        ValueError if version is found but not valid
    """
    v = getattr(mod, "__version__", None)
    if v is None:
        return None
    pat = r"\d+\.\d+\.\d+.*"
    if not re.match(pat, v):
        raise ValueError(
            'Version "{}" does not match regular expression '
            'pattern "{}"'.format(v, pat)
        )
    return v


def get_module_author(mod):
    """Find and return the module author.

    Args:
        mod (module): Python module
    Returns:
        (str) Author string or None if not found
    Raises:
        nothing
    """
    return getattr(mod, "__author__", None)


class TempDir(object):
    """Simple context manager for mkdtemp().
    """

    def __init__(self, *args):
        self._d = None
        self._a = args

    def __enter__(self):
        self._d = tempfile.mkdtemp(*self._a)
        return self._d

    def __exit__(self, *args):
        if self._d:
            shutil.rmtree(self._d)
            self._d = None


def is_jupyter_notebook(filename, check_contents=True):
    # type: (str) -> bool
    """See if this is a Jupyter notebook.
    """
    if not filename.endswith(".ipynb"):
        return False
    if check_contents:
        try:
            nb = json.load(open(filename))
        except (UnicodeDecodeError, JSONDecodeError):
            return False
        for key in "cells", "metadata", "nbformat":
            if key not in nb:
                return False
    return True


def is_python(filename):
    # type: (str) -> bool
    """See if this is a Python file.
    Do *not* import the source code.
    """
    if not filename.endswith(".py"):
        return False
    return True  # XXX: look inside?


def is_resource_json(filename, max_bytes=1e6):
    """Is this file a JSON Resource?

    Args:
        filename (str): Full path to file
        max_bytes (int): Max. allowable size. Since we try to parse
             the file, this saves potential DoS issues. Large files
             are a bad idea anyways, since this is metadata and may
             be stored somewhere with a record size limit (like MongoDB).

    Returns:
        (bool) Whether it's a resource JSON file.
    """
    if not filename.endswith(".json"):
        return False
    # get size
    st = os.stat(filename)
    # if it's under max_bytes, parse it
    if st.st_size <= max_bytes:
        try:
            d = json.load(open(filename))
        except (UnicodeDecodeError, JSONDecodeError):
            return False
        # look for a couple distinctive keys
        for key in "id_", "type":
            if key not in d:
                return False
        return True
    else:
        # if it's over max_bytes, it's "bad"
        return False


def datetime_timestamp(v):
    """Get numeric timestamp.
    This will work under both Python 2 and 3.

    Args:
        v (datetime.datetime): Date/time value

    Returns:
        (float) Floating point timestamp
    """
    if hasattr(v, "timestamp"):  # Python 2/3 test
        # Python 2
        result = v.timestamp()
    else:
        # Python 3
        result = time.mktime(v.timetuple()) + v.microsecond / 1e6
    return result


class ColorTerm:
    """For colorized printing, a very simple wrapper that
    allows colorama objects, or nothing, to be used.
    """

    class EmptyStr:
        """Return an empty string on any attribute requested."""

        def __getattr__(self, a):
            return ""

    def __init__(self, enabled=True):
        self._width = None
        if enabled:
            colorama.init(autoreset=True)
            # Colorama colors and styles
            F = self.Fore = colorama.Fore
            B = self.Back = colorama.Back
            S = self.Style = colorama.Style
            # Aliases for colors and styles
            (
                self.blue,
                self.green,
                self.yellow,
                self.red,
                self.cyan,
                self.magenta,
                self.bluebg,
                self.greenbg,
                self.yellowbg,
                self.redbg,
                self.cyanbg,
                self.magentabg,
                self.white,
                self.black,
                self.whitebg,
                self.blackbg,
                self.bold,
                self.dim,
                self.normal,
                self.reset,
                self.resetc,
            ) = (
                F.BLUE,
                F.GREEN,
                F.YELLOW,
                F.RED,
                F.CYAN,
                F.MAGENTA,
                B.BLUE,
                B.GREEN,
                B.YELLOW,
                B.RED,
                B.CYAN,
                B.MAGENTA,
                F.WHITE,
                F.BLACK,
                B.WHITE,
                B.BLACK,
                S.BRIGHT,
                S.DIM,
                S.NORMAL,
                S.RESET_ALL,
                F.RESET,
            )
        else:
            # Make any attribute (e.g. `Fore.BLUE`) return an empty
            # string, thus disabling all the color codes.
            self.Fore, self.Back, self.Style = (
                self.EmptyStr(),
                self.EmptyStr(),
                self.EmptyStr(),
            )
            (
                self.blue,
                self.green,
                self.yellow,
                self.red,
                self.cyan,
                self.magenta,
                self.bluebg,
                self.greenbg,
                self.yellowbg,
                self.redbg,
                self.cyanbg,
                self.magentabg,
                self.white,
                self.black,
                self.whitebg,
                self.blackbg,
                self.bold,
                self.dim,
                self.normal,
                self.reset,
                self.resetc,
            ) = [""] * 21

    @property
    def width(self):
        if self._width is None:
            self._width = shutil.get_terminal_size()[0]
        return self._width


def mkdir_p(path, *args):
    """Try to create all non-existent components of a path.

    Args:
        path (str): Path to create
        args: Other arguments for `os.mkdir()`.
    Returns:
        None
    Raises:
        os.error: Raised from `os.mkdir()`
    """
    plist = path.split(os.path.sep)
    if plist[0] == "":
        # do not try to create filesystem root
        dir_name = os.path.sep
        plist = plist[1:]
    else:
        dir_name = ""
    for p in plist:
        dir_name = os.path.join(dir_name, p)
        if not os.path.exists(dir_name):
            os.mkdir(dir_name, *args)


def uuid_prefix_len(uuids, step=4, maxlen=32):
    """Get smallest multiple of `step` len prefix that gives unique values.

    The algorithm is not fancy, but good enough: build *sets* of
    the ids at increasing prefix lengths until the set has all ids (no duplicates).
    Experimentally this takes ~.1ms for 1000 duplicate ids (the worst case).
    """
    full = set(uuids)
    all_of_them = len(full)
    for n in range(step, maxlen, step):
        prefixes = {u[:n] for u in uuids}
        if len(prefixes) == all_of_them:
            return n
    return maxlen



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################


def writethis(*arg):
    # This function is used to write to the terminal
    import sys
    for pnt in arg:
        if isinstance(pnt, str):
            sys.stdout.write(pnt)
        else:
            sys.stdout.write('!!! Printing Variable !!!')
            sys.stdout.write(pnt)



# -*- coding: UTF-8 -*-
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module contains utility functions for mathematical operators of use in
equation oriented models.
"""

from pyomo.environ import Param

__author__ = "Andrew Lee"


def smooth_abs(a, eps=1e-4):
    """General function for creating an expression for a smooth minimum or
    maximum.

    .. math:: |a| = sqrt(a^2 + eps^2)

    Args:
        a : term to get absolute value from (Pyomo component, float or int)
        eps : smoothing parameter (Param, float or int) (default=1e-4)

    Returns:
        An expression for the smoothed absolute value operation.
    """
    # Check type of eps
    if not (isinstance(eps, (float, int, Param))):
        raise TypeError("smooth_abs eps argument must be a float, int or "
                        "Pyomo Param")

    # Create expression
    try:
        expr = (a**2 + eps**2)**0.5
    except TypeError:
        raise TypeError("Unsupported argument type for smooth_abs. Must be "
                        "a Pyomo Var, Param or Expression, or a float or int.")

    return expr


def smooth_minmax(a, b, eps=1e-4, sense='max'):
    """General function for creating an expression for a smooth minimum or
    maximum. Uses the smooth_abs operator.

    .. math:: minmax(a, b) = 0.5*(a+b +- |a-b|)

    Args:
        a : first term in mix or max function (Pyomo component, float or int)
        b : second term in min or max function (Pyomo component, float or int)
        eps : smoothing parameter (Param, float or int) (default=1e-4)
        sense : 'mim' or 'max' (default = 'max')

    Returns:
        An expression for the smoothed minimum or maximum operation.
    """
    # Check type of eps
    if not (isinstance(eps, (float, int, Param))):
        raise TypeError("Smooth {} eps argument must be a float, int or "
                        "Pyomo Param".format(sense))

    # Set sense of expression
    if sense == 'max':
        mm = 1
    elif sense == 'min':
        mm = -1
    else:
        raise ValueError("Unrecognised sense argument to smooth_minmax. "
                         "Must be 'min' or 'max'.")

    # Create expression
    try:
        expr = 0.5*(a+b+mm*smooth_abs(a-b, eps))
    except TypeError:
        raise TypeError("Unsupported argument type for smooth_{}. Must be "
                        "a Pyomo Var, Param or Expression, or a float or int."
                        .format(sense))

    return expr


def smooth_max(a, b, eps=1e-4):
    """Smooth maximum operator, using smooth_abs operator.

    .. math:: max(a, b) = 0.5*(a+b + |a-b|)

    Args:
        a : first term in max function
        b : second term in max function
        eps : smoothing parameter (Param or float, default = 1e-4)

    Returns:
        An expression for the smoothed maximum operation.
    """
    expr = smooth_minmax(a, b, eps, sense='max')
    return expr


def smooth_min(a, b, eps=1e-4):
    """Smooth minimum operator, using smooth_abs operator.

    .. math:: max(a, b) = 0.5*(a+b - |a-b|)

    Args:
        a : first term in min function
        b : second term in min function
        eps : smoothing parameter (Param or float, default = 1e-4)

    Returns:
        An expression for the smoothed minimum operation.
    """
    expr = smooth_minmax(a, b, eps, sense='min')
    return expr



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Demonstration and test flowsheet for a dynamic flowsheet.

"""

import matplotlib.pyplot as plt

# Import Pyomo libraries
from pyomo.environ import (ConcreteModel, SolverFactory, TransformationFactory,
                           Var, Constraint)
from pyomo.network import Arc

# Import IDAES core
from idaes.core import FlowsheetBlock

# Import Unit Model Modules
import idaes.property_models.examples.saponification_thermo as thermo_props
import idaes.property_models.examples.saponification_reactions as \
    reaction_props

# Import Unit Model Modules
from idaes.unit_models import CSTR, Mixer


def main():
    """
    Make the flowsheet object, fix some variables, and solve the problem
    """
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Add a flowsheet object to the model
    # time_set has points at 0 and 20 as the start and end of the domain,
    # and a point at t=1 to allow for a step-change at this time
    m.fs = FlowsheetBlock(default={"dynamic": True,
                                   "time_set": [0, 1, 20]})

    # Add property packages to flowsheet library
    m.fs.thermo_params = thermo_props.SaponificationParameterBlock()
    m.fs.reaction_params = reaction_props.SaponificationReactionParameterBlock(
        default={"property_package": m.fs.thermo_params})

    # Create unit models
    m.fs.mix = Mixer(default={"dynamic": False,
                              "property_package": m.fs.thermo_params})
    m.fs.Tank1 = CSTR(default={"property_package": m.fs.thermo_params,
                               "reaction_package": m.fs.reaction_params,
                               "has_holdup": True,
                               "has_equilibrium_reactions": False,
                               "has_heat_of_reaction": True,
                               "has_heat_transfer": True,
                               "has_pressure_change": False})
    m.fs.Tank2 = CSTR(default={"property_package": m.fs.thermo_params,
                               "reaction_package": m.fs.reaction_params,
                               "has_holdup": True,
                               "has_equilibrium_reactions": False,
                               "has_heat_of_reaction": True,
                               "has_heat_transfer": True,
                               "has_pressure_change": False})

    # Add pressure-flow constraints to Tank 1
    m.fs.Tank1.height = Var(m.fs.time,
                            initialize=1.0,
                            doc="Depth of fluid in tank [m]")
    m.fs.Tank1.area = Var(initialize=1.0,
                          doc="Cross-sectional area of tank [m^2]")
    m.fs.Tank1.flow_coeff = Var(m.fs.time,
                                initialize=5e-5,
                                doc="Tank outlet flow coefficient")

    def geometry(b, t):
        return b.volume[t] == b.area*b.height[t]
    m.fs.Tank1.geometry = Constraint(m.fs.time, rule=geometry)

    def outlet_flowrate(b, t):
        return b.control_volume.properties_out[t].flow_vol == \
                    b.flow_coeff[t]*b.height[t]**0.5
    m.fs.Tank1.outlet_flowrate = Constraint(m.fs.time,
                                            rule=outlet_flowrate)

    # Add pressure-flow constraints to tank 2
    m.fs.Tank2.height = Var(m.fs.time,
                            initialize=1.0,
                            doc="Depth of fluid in tank [m]")
    m.fs.Tank2.area = Var(initialize=1.0,
                          doc="Cross-sectional area of tank [m^2]")
    m.fs.Tank2.flow_coeff = Var(m.fs.time,
                                initialize=5e-5,
                                doc="Tank outlet flow coefficient")

    m.fs.Tank2.geometry = Constraint(m.fs.time, rule=geometry)
    m.fs.Tank2.outlet_flowrate = Constraint(m.fs.time,
                                            rule=outlet_flowrate)

    # Make Streams to connect units
    m.fs.stream1 = Arc(source=m.fs.mix.outlet,
                       destination=m.fs.Tank1.inlet)

    m.fs.stream2 = Arc(source=m.fs.Tank1.outlet,
                       destination=m.fs.Tank2.inlet)



    # Discretize time domain
    m.discretizer = TransformationFactory('dae.finite_difference')
    m.discretizer.apply_to(m,
                           nfe=50,
                           wrt=m.fs.time,
                           scheme="BACKWARD")

    TransformationFactory("network.expand_arcs").apply_to(m)

    # Set inlet and operating conditions, and some initial conditions.
    m.fs.mix.inlet_1.flow_vol.fix(0.5)
    m.fs.mix.inlet_1.conc_mol_comp[:, "H2O"].fix(55388.0)
    m.fs.mix.inlet_1.conc_mol_comp[:, "NaOH"].fix(100.0)
    m.fs.mix.inlet_1.conc_mol_comp[:, "EthylAcetate"].fix(0.0)
    m.fs.mix.inlet_1.conc_mol_comp[:, "SodiumAcetate"].fix(0.0)
    m.fs.mix.inlet_1.conc_mol_comp[:, "Ethanol"].fix(0.0)
    m.fs.mix.inlet_1.temperature.fix(303.15)
    m.fs.mix.inlet_1.pressure.fix(101325.0)

    m.fs.mix.inlet_2.flow_vol.fix(0.5)
    m.fs.mix.inlet_2.conc_mol_comp[:, "H2O"].fix(55388.0)
    m.fs.mix.inlet_2.conc_mol_comp[:, "NaOH"].fix(0.0)
    m.fs.mix.inlet_2.conc_mol_comp[:, "EthylAcetate"].fix(100.0)
    m.fs.mix.inlet_2.conc_mol_comp[:, "SodiumAcetate"].fix(0.0)
    m.fs.mix.inlet_2.conc_mol_comp[:, "Ethanol"].fix(0.0)
    m.fs.mix.inlet_2.temperature.fix(303.15)
    m.fs.mix.inlet_2.pressure.fix(101325.0)

    m.fs.Tank1.area.fix(1.0)
    m.fs.Tank1.flow_coeff.fix(0.5)
    m.fs.Tank1.heat_duty.fix(0.0)

    m.fs.Tank2.area.fix(1.0)
    m.fs.Tank2.flow_coeff.fix(0.5)
    m.fs.Tank2.heat_duty.fix(0.0)

    # Set initial conditions - accumulation = 0 at time = 0
    m.fs.fix_initial_conditions(state="steady-state")

    # Initialize Units
    m.fs.mix.initialize()

    m.fs.Tank1.initialize(state_args={
            "flow_vol": 1.0,
            "conc_mol_comp": {"H2O": 55388.0,
                              "NaOH": 100.0,
                              "EthylAcetate": 100.0,
                              "SodiumAcetate": 0.0,
                              "Ethanol": 0.0},
            "temperature": 303.15,
            "pressure": 101325.0})

    m.fs.Tank2.initialize(state_args={
            "flow_vol": 1.0,
            "conc_mol_comp": {"H2O": 55388.0,
                              "NaOH": 100.0,
                              "EthylAcetate": 100.0,
                              "SodiumAcetate": 0.0,
                              "Ethanol": 0.0},
            "temperature": 303.15,
            "pressure": 101325.0})

    # Create a solver
    solver = SolverFactory('ipopt')
    results = solver.solve(m.fs)

    # Make a step disturbance in feed and solve again
    for t in m.fs.time:
        if t >= 1.0:
            m.fs.mix.inlet_2.conc_mol_comp[t, "EthylAcetate"].fix(90.0)
    results = solver.solve(m.fs)

    # Print results
    print(results)

    # For testing purposes
    return(m, results)


def plot_results(m):
    plt.figure("Tank 1 Outlet")
    plt.plot(m.fs.time,
             list(m.fs.Tank1.outlet.conc_mol_comp[:, "NaOH"].value),
             label='NaOH')
    plt.plot(m.fs.time,
             list(m.fs.Tank1.outlet.conc_mol_comp[:, "EthylAcetate"].value),
             label='EthylAcetate')
    plt.plot(m.fs.time,
             list(m.fs.Tank1.outlet.conc_mol_comp[:, "SodiumAcetate"].value),
             label='SodiumAcetate')
    plt.plot(m.fs.time,
             list(m.fs.Tank1.outlet.conc_mol_comp[:, "Ethanol"].value),
             label='Ethanol')
    plt.legend()
    plt.grid()
    plt.xlabel("Time [s]")
    plt.ylabel("Concentration [mol/m^3]")

    plt.figure("Tank 2 Outlet")
    plt.plot(m.fs.time,
             list(m.fs.Tank2.outlet.conc_mol_comp[:, "NaOH"].value),
             label='NaOH')
    plt.plot(m.fs.time,
             list(m.fs.Tank2.outlet.conc_mol_comp[:, "EthylAcetate"].value),
             label='EthylAcetate')
    plt.plot(m.fs.time,
             list(m.fs.Tank2.outlet.conc_mol_comp[:, "SodiumAcetate"].value),
             label='SodiumAcetate')
    plt.plot(m.fs.time,
             list(m.fs.Tank2.outlet.conc_mol_comp[:, "Ethanol"].value),
             label='Ethanol')
    plt.legend()
    plt.grid()
    plt.xlabel("Time [s]")
    plt.ylabel("Concentration [mol/m^3]")
    plt.show(block=True)


if __name__ == "__main__":
    m, results = main()
    plot_results(m)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# This subroutine is no longer used in the pyomo-ripe implementation


def read(infile):
    # Arguemnts
    # infile : ripe input file to be parsed
    import ripe
    import sys
    import numpy as np

    data, rspace = ripe.data, ripe.rspace
    # debug = ripe.debug

    # Hard coded delimiter
    dlm = " "

    # thigns that needs checks - mechansisms, clc, pclabels,
    # define options to be parsed automatically
    strlist = ["simulator"]
    intlist = ["ccon", "eobj", "npc", "npcval", "ntimedata", "timeseries"]
    llist = [
        "mechanisms",
        "clc",
        "slabels",
        "pclabels",
        "slabels",
        "pclabels",
        "noise",
        "adsspec",
        "adslabel",
        "revlabel",
        "revspec",
    ]
    nlist = ["savescratch", "c0vary", "volume"]
    mlist = ["begin_tdata", "begin_tvaldata", "begin_vdata"]
    endlist = ["end_tdata", "end_tvaldata", "end_vdata"]
    templist = ["T", "Tval", "v"]
    temp2list = ["npct", "npcval", "npcv"]
    labs = {}
    labs2 = {}
    for item, lab in zip(mlist, templist):
        labs[item] = lab
    for item, lab in zip(mlist, temp2list):
        labs2[item] = lab

    # open and read .ripe
    f = open(infile)
    f1 = f.read()
    f.close()
    rd = f1.split("\n")
    ln = 0
    while ln < len(rd):
        if rd[ln] == "":
            ln = ln + 1
            continue
        if rd[ln][0] in list(["%", "#"]):
            ln = ln + 1
            continue
        line = rd[ln].split(dlm)
        while "" in line:
            line.remove("")
        while " " in line:
            line.remove(" ")
        inopt = line[0]
        if inopt not in rspace["inopt"]:
            sys.stdout.write(
                "Error : I don't understand the input file at line : "
                + str(ln + 1)
                + "\n"
            )
            exit
        # check for string inputs
        elif inopt in strlist:
            data["otherdata"][inopt] = str(line[-1])
            sys.stdout.write(data["otherdata"][inopt])
        #            ln+=1
        #            continue
        elif inopt in llist:
            data["otherdata"][inopt] = line[1:]
        elif inopt in intlist:
            data["otherdata"][inopt] = int(line[-1])
        elif inopt in nlist:
            data["otherdata"][inopt] = line[-1]
        elif inopt in mlist:
            n = 1
            nline = rd[ln + n].split(dlm)
            while nline[0] not in endlist:
                try:
                    nline = rd[ln + n].split(dlm)
                except Exception:
                    nline = rd[ln + n]
                n += 1
            if n > 3:
                dlist = rd[ln + 1 : ln + n].split("\n")
                dlist = [s.split(dlm) for s in dlist]
            else:
                dlist = rd[ln + 1]
            data["procdata"][labs[inopt]] = np.asarray(dlist)
            data["procdata"][labs2[inopt]] = n
            ln = ln + n + 1
            continue

        # concdata is read outside of the loop
        elif inopt == "begin_concdata":
            # timeseries data is read differently from data indexed over process conditions
            if "timeseries" in data["otherdata"]:
                ln = readtimeseries(rd, ln, dlm, "procdata", "concdata")
                continue
            else:
                for s in data["otherdata"]["slabels"]:
                    data["concdata"][s] = np.zeros([data["otherdata"]["npc"]])
                t2 = 0
                for i in range(ln + 1, ln + 1 + data["otherdata"]["npc"]):
                    rd2 = rd[i].split(dlm)
                    while "" in rd2:
                        rd2.remove("")
                    while " " in rd2:
                        rd2.remove(" ")
                    t3 = 0
                    for s in data["otherdata"]["slabels"]:
                        data["concdata"][s][t2] = float(rd2[t3])
                        t3 = t3 + 1
                    t2 = t2 + 1
                ln = ln + data["otherdata"]["npc"] + 1
                continue

        # read validation conc data
        elif inopt == "begin_valdata":
            if "timeseries" in data["otherdata"]:
                ln = readtimeseries(rd, ln, dlm, "procvaldata", "concvaldata")
                continue
            else:
                for s in data["otherdata"]["slabels"]:
                    data["concvaldata"][s] = np.zeros([data["otherdata"]["npcval"]])
                t2 = 0
                for i in range(ln + 1, ln + 1 + data["otherdata"]["npcval"]):
                    rd2 = rd[i].split(dlm)
                    while "" in rd2:
                        rd2.remove("")
                    while " " in rd2:
                        rd2.remove(" ")
                    t3 = 0
                    for s in data["otherdata"]["slabels"]:
                        data["concvaldata"][s][t2] = float(rd2[t3])
                        t3 = t3 + 1
                    t2 = t2 + 1
                ln = ln + data["otherdata"]["npcval"] + 1
                continue
        # Read initial concentrations
        elif inopt == "begin_c0data":
            data["procdata"]["c0data"] = {}
            if "pclabels" in data["otherdata"]:
                if "x0" in data["otherdata"]["pclabels"]:
                    if "npc" in data["otherdata"]:
                        temp = int(data["otherdata"]["npc"])
                    else:
                        sys.stdout.write(
                            "Must specify number of process conditions before c0data"
                        )
                elif "T" in data["otherdata"]["pclabels"]:
                    if "npc" in data["otherdata"]:
                        temp = int(data["otherdata"]["npc"])
                    else:
                        sys.stdout.write(
                            "Must specify number of process conditions before c0data"
                        )
                else:
                    temp = 1
            else:
                temp = 1
            for s in data["otherdata"]["slabels"]:
                data["procdata"]["c0data"][s] = np.zeros(temp)
            for i in range(temp):
                rd2 = rd[ln + i + 1].split(dlm)
                while "" in rd2:
                    line.remove("")
                while " " in rd2:
                    line.remove(" ")
                t1 = 0
                for s in data["otherdata"]["slabels"]:
                    data["procdata"]["c0data"][s][i] = float(rd2[t1])
                    t1 = t1 + 1
            ln = ln + 1 + temp
            continue

        # Read custom adsorption ~constants(T)
        # Current ads schemes must be applied to all adsspec
        elif inopt == "begin_ads":
            if "adsspec" not in data["otherdata"]:
                sys.stdout.write("Define species involved in adsorption\n")
                exit
            elif "adslabel" not in data["otherdata"]:
                sys.stdout.write("Define labels for adsorption terms\n")
                exit
            # Initilize the data dictionary
            data["procdata"]["adata"] = {}
            for i in data["otherdata"]["adslabel"]:
                data["procdata"]["adata"][i] = {}
                for j in data["otherdata"]["slabels"]:
                    data["procdata"]["adata"][i][j] = ""
            t1 = 0
            for i in range(
                ln + 1,
                ln
                + len(data["otherdata"]["adslabel"]) * len(data["otherdata"]["adsspec"])
                + 1,
            ):
                data["procdata"]["adata"][data["otherdata"]["adslabel"][t1]][
                    data["otherdata"]["slabels"][i - ln - 1]
                ] = str(rd[i])
                if i - ln - i == len(data["otherdata"]["adsspec"]):
                    t1 = t1 + 1
            ln = (
                ln
                + 1
                + len(data["otherdata"]["adslabel"]) * len(data["otherdata"]["adsspec"])
            )
            # Take care of species not included in adsorption
            for s in [
                val
                for val in data["otherdata"]["slabels"]
                if val not in data["otherdata"]["adsspec"]
            ]:
                for r in data["otherdata"]["adslabel"]:
                    data["procdata"]["adata"][r][s] = "0.0"
            continue

        # Read eq constant(T)
        # 1 label per reaction > one line in data section
        elif inopt == "begin_rev":
            if "revspec" not in data["otherdata"]:
                sys.stdout.write("Define species involved in reversible reaction\n")
                exit
            elif "revlabel" not in data["otherdata"]:
                sys.stdout.write("Define labels for eq constants\n")
                exit
            # Initilize the data dictionary
            data["procdata"]["revdata"] = {}
            for i in data["otherdata"]["revlabel"]:
                data["procdata"]["revdata"][i] = {}
                for j in data["otherdata"]["revspec"]:
                    data["procdata"]["revdata"][i][j] = ""
            t1 = 0
            for i in range(ln + 1, ln + len(data["otherdata"]["revspec"]) + 1):
                data["procdata"]["revdata"][data["otherdata"]["revlabel"][0]][
                    data["otherdata"]["revspec"][i - ln - 1]
                ] = str(rd[i])
            ln = ln + 1 + len(data["otherdata"]["revspec"])
            continue

        # End of the loop
        ln = ln + 1
    # perform some checks
    if "ntimedata" not in data["otherdata"]:
        data["otherdata"]["ntimedata"] = 0


def readtimeseries(rd, ln, dlm, st1, st2):
    import ripe

    # Read time series data
    import numpy as np

    data, debug = ripe.data, ripe.debug

    data[st1]["time"] = np.array(
        np.zeros([data["otherdata"]["timeseries"], data["otherdata"]["ntimedata"]])
    )
    for s in data["otherdata"]["slabels"]:
        data[st2][s] = np.zeros(
            [data["otherdata"]["timeseries"], data["otherdata"]["ntimedata"]]
        )
        t1 = 0
        t2 = 0
        for i in range(
            ln + 1, ln + 1 + data["otherdata"]["ntimedata"] * data["otherdata"]["npc"]
        ):
            rd2 = rd[i].split(dlm)
            if len(rd2) == 1:
                rd2 = rd[i].split("\t")
                while "" in rd2:
                    rd2.remove("")
                while " " in rd2:
                    rd2.remove(" ")
                data["procdata"]["time"][t1][t2] = float(rd2[0])
                t3 = 0
                for s in data["otherdata"]["slabels"]:
                    # Check for true model specified to debug
                    if not debug["true"]["use"]:
                        data[st2][s][t1][t2] = float(rd2[t3 + 1])
                    else:
                        data[st2][s][t1][t2] = 1 - np.exp(
                            -debug["true"]["k"]
                            * np.exp(-debug["true"]["e"] / (8.314 * data[st1]["T"][t1]))
                            * data[st1]["time"][t1][t2]
                        )
                    t3 = t3 + 1
                t2 = t2 + 1
                if t2 == data["otherdata"]["ntimedata"]:
                    t1 = t1 + 1
                    t2 = 0
        ln = ln + data["otherdata"]["ntimedata"] * data["otherdata"]["timeseries"] + 3
        return ln



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

import numpy as np

# _all__ = ["lin", "linjac", "arr", "arrjac", "refarr", "refarrjac"]
# This subroutine contains definitions for the arrhenious form accepted by RIPE
# Currently, the forms cannot be defined by the user
# The lin,arr, and refarra subroutines return squared residuals
# the linjac, arrjac, and refarrjac subroutiens return the jocbian of the
# esitmated parameter


def lin(y, x, a):
    n, s, ln = np.shape(x)
    guess = 0
    out = np.zeros(n * s)
    q = 0
    for i in range(n):
        for j in range(s):
            guess = 0.0
            for k in range(ln):
                guess += a[k] * x[i, j, k]
            out[q] = (y[i, j] - guess) ** 2
            q += 1
    return out


def linjac(y, x, a):
    n, s, ln = np.shape(x)
    out = np.zeros([n * s, ln])
    q = 0
    for i in range(n):
        for j in range(s):
            for k in range(ln):
                out[q, k] = x[i, j, k]
            q += 1
    return out


def arr(y, T, x, p, gc):
    n, s, ln = np.shape(x)
    out = np.zeros(n * s)
    q = 0
    for i in range(n):
        for j in range(s):
            guess = 0.0
            for k in range(ln):
                guess += p[k] * np.exp((-p[k + ln] / (gc * T[i]))) * x[i, j, k]
            out[q] = (y[i, j] - guess) ** 2  # / sig[i,j]
            q += 1
    return out


def arrjac(y, T, x, p, gc):
    n, s, ln = np.shape(x)
    out = np.zeros([n * s, 2 * ln])
    q = 0
    for i in range(n):
        for j in range(s):
            for k in range(ln):
                out[q, k] = np.exp((-p[k + ln] / (gc * T[i]))) * x[i, j, k]
                out[q, k + ln] = (
                    p[k]
                    * (-1 / (gc * T[i]))
                    * np.exp((-p[k + ln] / (gc * T[i])))
                    * x[i, j, k]
                )
            q += 1
    return out


def refarr(y, T, Tref, x, p, gc):
    n, s, ln = np.shape(x)
    out = np.zeros(n * s)
    q = 0
    for i in range(n):
        for j in range(s):
            guess = 0.0
            for k in range(ln):
                guess += (
                    p[k]
                    * np.exp(-1.0 * p[k + ln] / (gc) * (1 / (T[i]) - 1 / (Tref)))
                    * x[i, j, k]
                )
            out[q] = (y[i, j] - guess) ** 2  # / sig[i,j]
            q += 1
    return out


def refarrjac(y, T, Tref, x, p, gc):
    n, s, ln = np.shape(x)
    out = np.zeros([n * s, 2 * ln])
    q = 0
    for i in range(n):
        for j in range(s):
            for k in range(ln):
                out[q, k] = (
                    np.exp(((-1.0 * p[k + ln]) / (gc)) * (1 / (T[i]) - 1 / (Tref)))
                    * x[i, j, k]
                )
                out[q, k + ln] = (
                    p[k]
                    * ((-1.0 / (gc)) * (1 / (T[i]) - 1 / (Tref)))
                    * np.exp(((-1 * p[k + ln]) / (gc)) * ((1 / (T[i])) - (1 / (Tref))))
                    * x[i, j, k]
                )
            q += 1
    return out



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Example property package for the VLE calucations for a Benzene-Toluene-o-Xylene
system. If using the activity coefficient models (NRTL or Wilson), the user is
expected to provide the paramters necessary for these models. Please note that
these parameters are declared as variables here to allow for use in a parameter
estimation problem if the VLE data is available.
"""

# Import Pyomo libraries
from pyomo.environ import Param, NonNegativeReals, Set

# Import IDAES cores
from idaes.core import declare_process_block_class
from idaes.core.util.misc import extract_data

from idaes.property_models.activity_coeff_models.activity_coeff_prop_pack \
    import ActivityCoeffParameterData
from idaes.logger import getIdaesLogger

# Some more inforation about this module
__author__ = "Jaffer Ghouse"
__version__ = "0.0.1"


# Set up logger
_log = getIdaesLogger(__name__)


@declare_process_block_class("BTXParameterBlock")
class BTXParameterData(ActivityCoeffParameterData):

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(BTXParameterData, self).build()

        self.component_list_master = Set(initialize=['benzene',
                                                     'toluene',
                                                     'o-xylene'])

        # Component list - a list of component identifiers
        # NOTE: User needs to update this list; can be a subset or
        # equal to the master component list
        self.component_list = Set(initialize=['benzene', 'toluene'])

        # List of components in each phase (optional)
        self.phase_comp = {"Liq": self.component_list,
                           "Vap": self.component_list}

        # List of phase equilibrium index
        self.phase_equilibrium_idx_master = Set(initialize=[1, 2, 3])

        self.phase_equilibrium_idx = Set(initialize=[1, 2])

        self.phase_equilibrium_list_master = \
            {1: ["benzene", ("Vap", "Liq")],
             2: ["toluene", ("Vap", "Liq")],
             3: ["o-xylene", ("Vap", "Liq")]}

        self.phase_equilibrium_list = \
            {1: ["benzene", ("Vap", "Liq")],
             2: ["toluene", ("Vap", "Liq")]}

        # Thermodynamic reference state
        self.pressure_reference = Param(mutable=True,
                                        default=101325,
                                        doc='Reference pressure [Pa]')
        self.temperature_reference = Param(mutable=True,
                                           default=298.15,
                                           doc='Reference temperature [K]')

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        pressure_critical_data = {'benzene': 48.9e5,
                                  'toluene': 41e5,
                                  'o-xylene': 37.3e5
                                  }

        self.pressure_critical = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(pressure_critical_data),
            doc='Critical pressure [Pa]')

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        temperature_critical_data = {'benzene': 562.2,
                                     'toluene': 591.8,
                                     'o-xylene': 630.3
                                     }

        self.temperature_critical = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(temperature_critical_data),
            doc='Critical temperature [K]')

        # Gas Constant
        self.gas_const = Param(within=NonNegativeReals,
                               mutable=False,
                               default=8.314,
                               doc='Gas Constant [J/mol.K]')

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        mw_comp_data = {'benzene': 78.1136E-3,
                        'toluene': 92.1405E-3,
                        'o-xylene': 106.167e-3}

        self.mw_comp = Param(self.component_list,
                             mutable=False,
                             initialize=extract_data(mw_comp_data),
                             doc="molecular weight Kg/mol")

        # Constants for specific heat capacity, enthalpy
        # Sources: The Properties of Gases and Liquids (1987)
        #         4th edition, Chemical Engineering Series - Robert C. Reid
        #         Perry's Chemical Engineers Handbook
        #         - Robert H. Perry (Cp_liq)
        CpIG_data = {('Liq', 'benzene', 'A'): 1.29E5,
                     ('Liq', 'benzene', 'B'): -1.7E2,
                     ('Liq', 'benzene', 'C'): 6.48E-1,
                     ('Liq', 'benzene', 'D'): 0,
                     ('Liq', 'benzene', 'E'): 0,
                     ('Vap', 'benzene', 'A'): -3.392E1,
                     ('Vap', 'benzene', 'B'): 4.739E-1,
                     ('Vap', 'benzene', 'C'): -3.017E-4,
                     ('Vap', 'benzene', 'D'): 7.130E-8,
                     ('Vap', 'benzene', 'E'): 0,
                     ('Liq', 'toluene', 'A'): 1.40E5,
                     ('Liq', 'toluene', 'B'): -1.52E2,
                     ('Liq', 'toluene', 'C'): 6.95E-1,
                     ('Liq', 'toluene', 'D'): 0,
                     ('Liq', 'toluene', 'E'): 0,
                     ('Vap', 'toluene', 'A'): -2.435E1,
                     ('Vap', 'toluene', 'B'): 5.125E-1,
                     ('Vap', 'toluene', 'C'): -2.765E-4,
                     ('Vap', 'toluene', 'D'): 4.911E-8,
                     ('Vap', 'toluene', 'E'): 0,
                     ('Liq', 'o-xylene', 'A'): 3.65e4,
                     ('Liq', 'o-xylene', 'B'): 1.0175e3,
                     ('Liq', 'o-xylene', 'C'): -2.63,
                     ('Liq', 'o-xylene', 'D'): 3.02e-3,
                     ('Liq', 'o-xylene', 'E'): 0,
                     ('Vap', 'o-xylene', 'A'): -1.585e-1,
                     ('Vap', 'o-xylene', 'B'): 5.962e-1,
                     ('Vap', 'o-xylene', 'C'): -3.443e-4,
                     ('Vap', 'o-xylene', 'D'): 7.528E-8,
                     ('Vap', 'o-xylene', 'E'): 0}

        self.CpIG = Param(self.phase_list, self.component_list,
                          ['A', 'B', 'C', 'D', 'E'],
                          mutable=False,
                          initialize=extract_data(CpIG_data),
                          doc="parameters to compute Cp_comp")

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        pressure_sat_coeff_data = {('benzene', 'A'): -6.98273,
                                   ('benzene', 'B'): 1.33213,
                                   ('benzene', 'C'): -2.62863,
                                   ('benzene', 'D'): -3.33399,
                                   ('toluene', 'A'): -7.28607,
                                   ('toluene', 'B'): 1.38091,
                                   ('toluene', 'C'): -2.83433,
                                   ('toluene', 'D'): -2.79168,
                                   ('o-xylene', 'A'): -7.53357,
                                   ('o-xylene', 'B'): 1.40968,
                                   ('o-xylene', 'C'): -3.10985,
                                   ('o-xylene', 'D'): -2.85992}

        self.pressure_sat_coeff = Param(
            self.component_list,
            ['A', 'B', 'C', 'D'],
            mutable=False,
            initialize=extract_data(pressure_sat_coeff_data),
            doc="parameters to compute Cp_comp")

        # Standard heats of formation
        # Source: NIST Webbook, https://webbook.nist.gov
        # Retrieved 25th September 2019
        dh_form_data = {('Vap', 'benzene'): 82.9e3,
                        ('Vap', 'toluene'): 50.1e3,
                        ('Vap', 'o-xylene'): 19.0e3,
                        ('Liq', 'benzene'): 49.0e3,
                        ('Liq', 'toluene'): 12.0e3,
                        ('Liq', 'o-xylene'): -24.4e3}

        self.dh_form = Param(self.phase_list,
                             self.component_list,
                             mutable=False,
                             initialize=extract_data(dh_form_data),
                             doc="Standard heats of formation [J/mol]")

        # Standard entropy of formation
        # Source: Engineering Toolbox, https://www.engineeringtoolbox.com
        # o-xylene from NIST Webbook, https://webbook.nist.gov
        # Retrieved 9th October, 2019
        ds_form_data = {('Vap', 'benzene'): -269,
                        ('Vap', 'toluene'): -321,
                        ('Vap', 'o-xylene'): -353.6,
                        ('Liq', 'benzene'): -173,
                        ('Liq', 'toluene'): -220,
                        ('Liq', 'o-xylene'): -246}

        self.ds_form = Param(self.phase_list,
                             self.component_list,
                             mutable=False,
                             initialize=extract_data(ds_form_data),
                             doc="Standard entropy of formation [J/mol.K]")
